"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.tasks
    :platform: Unix
    :synopsis: Burp-UI tasks module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os

from flask import current_app
from datetime import timedelta, datetime
from celery.schedules import crontab
from celery.utils.log import get_task_logger
from time import gmtime, strftime, sleep

from burpui._compat import to_unicode  # noqa
from burpui.config import config  # noqa
from burpui.ext.tasks import celery  # noqa
from burpui.ext.cache import cache  # noqa
from burpui.sessions import session_manager  # noqa
from burpui.engines.server import BUIServer  # noqa
from burpui.exceptions import BUIserverException, TooManyRecordsException  # noqa
from burpui.api.client import ClientTreeAll  # noqa
from burpui.utils import NOTIF_ERROR

try:
    from burpui.ext.ws import socketio

    WS_AVAILABLE = True
except ImportError:
    WS_AVAILABLE = False

if config.get("WITH_SQL"):
    from burpui.ext.sql import db
else:
    db = None

bui = current_app  # type: BUIServer
logger = get_task_logger(__name__)
ME = __name__

LOCK_EXPIRE = 60 * 30  # Lock expires in 30 minutes

BEAT_SCHEDULE = {
    "backup-running-4-minutely": {
        "task": "{}.backup_running".format(ME),
        "schedule": timedelta(seconds=15),  # run every 15 seconds
    },
    "get-all-backups-every-twenty-minutes": {
        "task": "{}.get_all_backups".format(ME),
        "schedule": crontab(minute="*/20"),  # every 20 minutes
    },
    "get-all-clients-reports-every-twenty-minutes": {
        "task": "{}.get_all_clients_reports".format(ME),
        "schedule": crontab(minute="*/20"),  # every 20 minutes
    },
    "cleanup-expired-sessions-every-four-hours": {
        "task": "{}.cleanup_expired_sessions".format(ME),
        "schedule": crontab(minute=1, hour="*/4"),  # every four hours
    },
}

if db:
    from burpui.models import Task

    BEAT_SCHEDULE.update(
        {
            "cleanup-restore-hourly": {
                "task": "{}.cleanup_restore".format(ME),
                "schedule": crontab(minute="12"),  # run every hour
            },
        }
    )

if "CELERYBEAT_SCHEDULE" in celery.conf and isinstance(
    celery.conf["CELERYBEAT_SCHEDULE"], dict
):
    celery.conf["CELERYBEAT_SCHEDULE"].update(BEAT_SCHEDULE)
else:
    celery.conf["CELERYBEAT_SCHEDULE"] = BEAT_SCHEDULE


def acquire_lock(name, value="nyan", timeout=LOCK_EXPIRE):
    """Utility function to acquire a lock before processing the request"""
    lock = cache.cache.get(name)
    if lock:
        acquire_lock.lock = lock
        return False
    return cache.cache.add(name, value, timeout)


acquire_lock.lock = None


def release_lock(name):
    """Utility function to release a lock"""
    return cache.cache.delete(name)


def wait_for(lock_name, value, wait=10, timeout=LOCK_EXPIRE):
    """Utility function to wait until the given lock has been released"""
    old_lock = None
    if not acquire_lock(lock_name, value, timeout):
        logger.warn(
            "A task is already running. Wait for it: {}/{}".format(
                lock_name, acquire_lock.lock
            )
        )
        old_lock = acquire_lock.lock
        # The lock should be released after LOCK_EXPIRE max
        while not acquire_lock(lock_name, value, timeout):
            sleep(wait)

        # TODO: maybe we should check the status of task referenced by"old_lock"
        # to make sure the lock did not expire and the task is actually over
        logger.debug("lock released by: {}".format(old_lock))

    return old_lock


@celery.task(bind=True, ignore_result=True)
def backup_running(self):
    # run one at the time, if one task was already running, we just discard
    # the new attempt
    if not acquire_lock(self.name):
        return None
    try:
        res = bui.client.is_one_backup_running()
        cache.cache.set("backup_running_result", res, 120)
        if WS_AVAILABLE:
            running = False
            if isinstance(res, dict):
                for run in res.values():
                    if len(run) > 0:
                        running = True
                        break
            elif len(res) > 0:
                running = True
            socketio.emit("backup_running", running, namespace="/ws")
    finally:
        release_lock(self.name)


@celery.task(bind=True, ignore_result=True)
def get_all_backups(self):
    # run one at the time, if one task was already running, we just discard
    # the new attempt
    if not acquire_lock(self.name):
        return None
    try:
        backups = {}
        if bui.config["STANDALONE"]:
            for cli in bui.client.get_all_clients():
                backups[cli["name"]] = bui.client.get_client(cli["name"])
        else:
            for serv in bui.client.servers:
                backups[serv] = {}
                for cli in bui.client.get_all_clients(agent=serv):
                    backups[serv][cli["name"]] = bui.client.get_client(
                        cli["name"], agent=serv
                    )
        cache.cache.set("all_backups", backups, 3600)
    finally:
        release_lock(self.name)


@celery.task(bind=True, ignore_result=True)
def get_all_clients_reports(self):
    # run one at the time, if one task was already running, we just discard
    # the new attempt
    if not acquire_lock(self.name):
        return None
    try:
        reports = {}
        if bui.config["STANDALONE"]:
            reports = bui.client.get_clients_report(bui.client.get_all_clients())
        else:
            for serv in bui.client.servers:
                reports[serv] = bui.client.get_clients_report(
                    bui.client.get_all_clients(agent=serv), serv
                )
        cache.cache.set("all_clients_reports", reports, 3600)
    finally:
        release_lock(self.name)


@celery.task(
    ignore_result=True,
    max_retries=5,
    autoretry_for=(TooManyRecordsException,),
    retry_backoff=4,
)
def cleanup_expired_sessions():
    bucket = []

    def expires(sess):
        ret = session_manager.invalidate_session_by_id(sess.uuid)
        if ret:
            bucket.append(sess.uuid)
        return ret

    # remove expired sessions, limit to 10000 per batch
    list(map(expires, session_manager.get_expired_sessions(10000)))
    session_manager.bulk_session_delete_by_id(bucket)

    # if we still have more than 10000 expired session, schedule a new cleanup soon
    # unless we already ran 5 successive cleanups (in which case we will just wait
    # for the next schedule to trigger the task).
    if session_manager.get_expired_sessions_count() >= 10000:
        # we raise an exception so celery knows it has to restart this tasks
        # anytime soon.
        raise TooManyRecordsException


@celery.task(
    ignore_result=True,
    max_retries=3,
    autoretry_for=(TooManyRecordsException,),
    retry_backoff=True,
)
def cleanup_restore():
    bucket = []
    query = Task.query.filter(
        Task.task == "perform_restore", Task.expire <= datetime.utcnow()
    )
    for rec in query.limit(100):
        logger.info("Task expired: {}".format(rec))
        task = perform_restore.AsyncResult(rec.uuid)
        try:
            if task.state != "SUCCESS":
                logger.warn(
                    "Task is not done yet or did not end "
                    "successfully: {}".format(task.state)
                )
                continue
            if not task.result:
                logger.warn("The task did not return anything")
                continue
            server = task.result.get("server")
            path = task.result.get("path")
            if path:
                if server:
                    if not bui.client.del_file(path, agent=server):
                        logger.warn("'{}' already removed".format(path))
                else:
                    if os.path.isfile(path):
                        os.unlink(path)
        finally:
            bucket.append(rec.uuid)
            task.revoke()
    try:
        Task.query.filter(Task.uuid.in_(bucket)).delete(synchronize_session=False)
        db.session.commit()
    except:
        db.session.rollback()
    if query.count() > 100:
        raise TooManyRecordsException


@celery.task(bind=True)
def perform_restore(
    self,
    client,
    backup,
    files,
    strip,
    fmt,
    passwd,
    server=None,
    user=None,
    admin=False,
    room=None,
    expire=timedelta(minutes=60),
):
    ret = None
    # we can have only one restore per server-client at the time
    lock_name = "{}-{}-{}".format(self.name, server, client)

    # TODO: maybe do something with old_lock someday
    wait_for(lock_name, self.request.id)

    try:
        if server:
            filename = "restoration_%d_%s_on_%s_at_%s.%s" % (
                backup,
                client,
                server,
                strftime("%Y-%m-%d_%H_%M_%S", gmtime()),
                fmt,
            )
        else:
            filename = "restoration_%d_%s_at_%s.%s" % (
                backup,
                client,
                strftime("%Y-%m-%d_%H_%M_%S", gmtime()),
                fmt,
            )

        self.update_state(state="STARTED", meta={"step": "doing"})
        archive, err = bui.client.restore_files(
            client, backup, files, strip, fmt, passwd, server
        )
        if not archive:
            if not err:
                err = "Something went wrong while restoring"
            self.update_state(state="FAILURE", meta={"error": to_unicode(err)})
            logger.error("FAILURE: {}".format(err))
        else:
            ret = {
                "filename": filename,
                "path": to_unicode(archive),
                "user": user,
                "server": server,
                "admin": admin,
            }
            logger.debug(ret)

    finally:
        release_lock(lock_name)

    if db:
        curr = Task.query.filter_by(uuid=self.request.id).first()
        if curr:
            curr.expire = datetime.utcnow() + expire
            try:
                db.session.commit()
            except:
                db.session.rollback()

    if err:
        # make the task crash
        raise Exception(err)

    return ret


@celery.task(bind=True)
def delete_client(
    self, client, keepconf, delcert, revoke, template, delete, server, user
):
    parser = bui.client.get_parser(agent=server)
    self.update_state(state="STARTED", meta={"step": "doing"})

    res = parser.remove_client(client, keepconf, delcert, revoke, template, delete)
    if any(x == NOTIF_ERROR for x, _ in res):
        self.update_state(state="FAILURE", meta={"error": res})
        raise Exception(res)

    ret = {
        "result": res,
        "client": client,
        "server": server,
        "user": user,
        "kwargs": {
            "keepconf": keepconf,
            "delcert": delcert,
            "revoke": revoke,
            "template": template,
            "delete": delete,
            "template": template,
        },
    }
    logger.debug(ret)
    return ret


@celery.task(bind=True)
def load_all_tree(self, client, backup, server=None, user=None):
    key = "load_all_tree-{}-{}-{}".format(client, backup, server)
    ret = cache.cache.get(key)
    if ret:
        return {
            "client": client,
            "backup": backup,
            "server": server,
            "user": user,
            "tree": ret,
        }

    lock_name = "{}-{}".format(self.name, server)

    # TODO: maybe do something with old_lock someday
    wait_for(lock_name, self.request.id)

    try:
        ret = ClientTreeAll._get_tree_all(client, backup, server)
    except BUIserverException as exp:
        raise Exception(str(exp))
    finally:
        release_lock(lock_name)

    cache.cache.set(key, ret, 3600)
    return {
        "client": client,
        "backup": backup,
        "server": server,
        "user": user,
        "tree": ret,
    }


def force_scheduling_now():
    """Force scheduling some tasks now"""
    get_all_backups.delay()
    backup_running.delay()
    get_all_clients_reports.delay()
    cleanup_expired_sessions.delay()
    cleanup_restore.delay()



# -*- coding: utf8 -*-
from .meta import meta_grants, BUIaclGrant
from .interface import BUIaclLoader
from ...utils import NOTIF_OK, NOTIF_WARN, NOTIF_ERROR, __


class ACLloader(BUIaclLoader):
    __doc__ = __("Uses the Burp-UI configuration file to load its rules.")
    section = name = "BASIC:ACL"
    priority = 100

    def __init__(self, app=None):
        """See :func:`burpui.misc.acl.interface.BUIaclLoader.__init__`

        :param app: Application context
        :type app: :class:`burpui.engines.server.BUIServer`
        """
        self.app = app
        self.conf = self.app.conf
        self.admins = ["admin"]
        self.moderators = []
        self._groups = {}
        self._grants = {}
        self.first_setup = True
        self.moderator = {}
        self._acl = meta_grants
        self.conf_id = None
        self.meta_id = meta_grants.id
        meta_grants.register_backend(self.name, self)
        self.load_acl(True)

    def reload(self):
        self.load_acl(True)

    def load_acl(self, force=False):
        if not force and self.conf_id:
            if not self.conf.changed(self.conf_id):
                return False

        # our config changed or we were forced to reload our rules.
        # if the meta_grants didn't change, we reset them
        # if they changed, it means something else triggered a reset
        if not meta_grants.changed(self.meta_id) and not self.first_setup:
            meta_grants.reset(self.name)

        self.first_setup = False

        self.admins = ["admin"]
        self.moderators = []
        self._grants = {}
        self.groups_def = {}

        adms = []
        mods = []

        def is_empty(data):
            return not data or data == [None]

        if self.section in self.conf.options:
            self.priority = self.conf.safe_get(
                "priority", "integer", section=self.section, defaults=self.priority
            )
            adms = self.conf.safe_get("admin", "force_list", section=self.section)
            mods = self.conf.safe_get("+moderator", "force_list", section=self.section)
            default_moderator = None
            if self.conf.get("STANDALONE"):
                default_moderator = '{"agents": {"rw": "local"}}'
            self.moderator = (
                self.conf.safe_get(
                    "@moderator",
                    "force_string",
                    section=self.section,
                    defaults=default_moderator,
                )
                or {}
            )
            meta_grants.set_moderator_grants(self.moderator)
            for opt in self.conf.options.get(self.section).keys():
                if opt in ["admin", "+moderator", "priority", "@moderator"]:
                    continue
                record = self.conf.safe_get(opt, "force_string", section=self.section)

                self.logger.debug("record: {} -> {}".format(opt, record))

                def _record(key):
                    if gname not in self.groups_def:
                        self.groups_def[gname] = {}
                    self.groups_def[gname][key] = parsed

                    return parsed

                if opt[0] == "+":
                    short = opt.lstrip("+")
                    gname = "@{}".format(short)
                    parsed = meta_grants.set_group(gname, record)
                    self._groups[short] = parsed
                    _record("members")
                elif opt[0] == "@":
                    short = opt.lstrip("@")
                    if short not in self._groups:
                        self._groups[short] = []
                    gname = opt
                    parsed = record
                    _record("grants")
                    meta_grants.set_grant(gname, parsed)
                else:
                    self._grants[opt] = BUIaclGrant(opt, record).grants
                    meta_grants.set_grant(opt, record)

        if not is_empty(adms):
            self.admins = adms
        if not is_empty(mods):
            self.moderators = mods

        meta_grants.set_admin(self.admins)
        meta_grants.set_moderator(self.moderators)

        self.logger.debug("admins: {}".format(self.admins))
        self.logger.debug("moderators: {}".format(self.moderators))
        self.logger.debug("moderator grants: {}".format(self.moderator))
        self.logger.debug("groups: {}".format(self.groups_def))

        self.conf_id = self.conf.id
        self.meta_id = meta_grants.id

        return True

    def _setup_acl(self):
        """Setup ACL management"""
        if not self.conf.lookup_section(self.section):
            self.conf._refresh()

    def add_grant(self, name, content):
        """Add a grant"""
        if name[0] in ["+", "@"]:
            message = "'{}' is not a valid grant name".format(name)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        self._setup_acl()
        if name in self.conf.options[self.section]:
            message = "grant '{}' already exists".format(name)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self.conf.options[self.section][name] = content
        self.conf.options.write()
        self.load_acl(True)
        message = "grant '{}' successfully added".format(name)
        return True, message, NOTIF_OK

    def del_grant(self, name):
        """Delete a grant"""
        if name[0] in ["+", "@"]:
            message = "'{}' is not a valid grant name".format(name)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        self._setup_acl()
        self.load_acl(True)
        if name not in self.conf.options[self.section]:
            message = "grant '{}' does not exist".format(name)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        del self.conf.options[self.section][name]
        self.conf.options.write()
        self.load_acl(True)
        if name in self.admins:
            self.del_admin(name)
        if name in self.moderators:
            self.del_moderator(name)
        for group, members in self._groups.items():
            if name in members:
                self.del_group_member(group, name)
        message = "grant '{}' successfully removed".format(name)
        return True, message, NOTIF_OK

    def mod_grant(self, name, content):
        """Update a grant"""
        if name[0] in ["+", "@"]:
            message = "'{}' is not a valid grant name".format(name)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        self._setup_acl()
        self.load_acl(True)
        if name not in self.conf.options[self.section]:
            message = "grant '{}' does not exist".format(name)
            self.logger.error(message)
            return False, message, NOTIF_WARN
        self.conf.options[self.section][name] = content
        self.conf.options.write()
        self.load_acl(True)
        message = "grant '{}' successfully modified".format(name)
        return True, message, NOTIF_OK

    def add_group(self, name, content):
        """Create a group"""
        self._setup_acl()
        name = "@{}".format(name)
        if name == "moderator":
            message = "'moderator' is a reserved name"
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        if name in self.conf.options[self.section]:
            message = "group '{}' already exists".format(name)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self.conf.options[self.section][name] = content
        self.conf.options.write()
        self.load_acl(True)
        message = "group '{}' successfully added".format(name)
        return True, message, NOTIF_OK

    def del_group(self, name):
        """Delete a group"""
        self._setup_acl()
        self.load_acl(True)
        gname = "@{}".format(name)
        if name == "moderator":
            message = "'moderator' is a reserved name"
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        if gname not in self.conf.options[self.section]:
            message = "group '{}' does not exist".format(name)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        del self.conf.options[self.section][gname]
        gmembers = "+{}".format(name)
        if gmembers in self.conf.options[self.section]:
            del self.conf.options[self.section][gmembers]
        self.conf.options.write()
        self.load_acl(True)
        if gname in self.admins:
            self.del_admin(gname)
        if gname in self.moderators:
            self.del_moderator(gname)
        for group, members in self._groups.items():
            if gname in members:
                self.del_group_member(group, gname)
        message = "grant '{}' successfully removed".format(name)
        return True, message, NOTIF_OK

    def mod_group(self, name, content):
        """Update a group"""
        self._setup_acl()
        name = "@{}".format(name)
        if name == "moderator":
            message = "'moderator' is a reserved name"
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        if name not in self.conf.options[self.section]:
            message = "group '{}' does not exist".format(name)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self.conf.options[self.section][name] = content
        self.conf.options.write()
        self.load_acl(True)
        message = "group '{}' successfully modified".format(name)
        return True, message, NOTIF_OK

    def add_group_member(self, group, member):
        """Add a user to a group"""
        if group not in self._groups:
            message = "group '{}' does not exist".format(group)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        if member in self._groups[group]:
            message = "'{}' already in group '{}'".format(member, group)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self._groups[group].append(member)
        gmembers = "+{}".format(group)
        self.conf.options[self.section][gmembers] = self._groups[group]
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' added to group '{}'".format(member, group)
        return True, message, NOTIF_OK

    def del_group_member(self, group, member):
        """Remove a user from a group"""
        if group not in self._groups:
            message = "group '{}' does not exist".format(group)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        if member not in self._groups[group]:
            message = "'{}' not in group '{}'".format(member, group)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self._groups[group].remove(member)
        gmembers = "+{}".format(group)
        self.conf.options[self.section][gmembers] = self._groups[group] or ""
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' removed from group '{}'".format(member, group)
        return True, message, NOTIF_OK

    def add_moderator(self, member):
        """Add a moderator"""
        if member in self.moderators:
            message = "'{}' is already a moderator".format(member)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self.moderators.append(member)
        self.conf.options[self.section]["+moderator"] = self.moderators
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' successfully added as moderator".format(member)
        return True, message, NOTIF_OK

    def del_moderator(self, member):
        """Delete a moderator"""
        if member not in self.moderators:
            message = "'{}' is not a moderator".format(member)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self.moderators.remove(member)
        self.conf.options[self.section]["+moderator"] = self.moderators or ""
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' successfully removed from moderators".format(member)
        return True, message, NOTIF_OK

    def mod_moderator(self, grants):
        """Update moderator grants"""
        self._setup_acl()
        self.moderator = grants
        self.conf.options[self.section]["@moderator"] = grants
        self.conf.options.write()
        self.load_acl(True)
        message = "moderator grants updated"
        return True, message, NOTIF_OK

    def add_admin(self, member):
        """Add an admin"""
        if member in self.admins:
            message = "'{}' is already an admin".format(member)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self.admins.append(member)
        self.conf.options[self.section]["admin"] = self.admins
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' successfully added as admin".format(member)
        return True, message, NOTIF_OK

    def del_admin(self, member):
        """Delete an admin"""
        if member not in self.admins:
            message = "'{}' is not an admin".format(member)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        self._setup_acl()
        self.admins.remove(member)
        self.conf.options[self.section]["admin"] = self.admins or ""
        self.conf.options.write()
        self.load_acl(True)
        message = "'{}' successfully removed from admins".format(member)
        return True, message, NOTIF_OK

    @property
    def acl(self):
        """Property to retrieve the backend"""
        if self._acl:
            self.load_acl()
            return self._acl
        return None  # pragma: no cover

    @property
    def grants(self):
        """Property to retrieve the list of grants"""
        return self._grants

    @property
    def groups(self):
        """Property to retrieve the list of groups with their members"""
        return self.groups_def



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.tasks
    :platform: Unix
    :synopsis: Burp-UI tasks api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import select
import struct

from . import api, cache_key, force_refresh
from .misc import History
from .custom import Resource
from .client import node_fields
from .clients import RunningBackup, ClientsReport, RunningClients
from ..engines.server import BUIServer  # noqa
from ..ext.cache import cache
from ..config import config
from ..decorators import browser_cache
from ..tasks import perform_restore, load_all_tree, delete_client, force_scheduling_now

from time import time
from zlib import adler32
from flask import (
    url_for,
    Response,
    current_app,
    after_this_request,
    send_file,
    request,
    g,
    session,
)
from flask_babel import gettext as _
from flask_restx import inputs
from flask_login import current_user
from datetime import timedelta
from werkzeug.datastructures import Headers

try:
    from .ext.ws import socketio  # noqa

    WS_AVAILABLE = True
except ImportError:
    WS_AVAILABLE = False

if config.get("WITH_SQL"):
    from ..ext.sql import db
    from ..models import Task
else:
    db = None

bui = current_app  # type: BUIServer
ns = api.namespace("tasks", "Asynchronous tasks methods")

# tuple composed with <task>, <callback url>
task_types = {
    "restore": (perform_restore, ".task_get_file"),
    "browse": (load_all_tree, ".task_do_browse_all"),
    "delete": (delete_client, ".task_deleted_client"),
}


@ns.route(
    "/status/<task_type>/<task_id>",
    "/<server>/status/<task_type>/<task_id>",
    endpoint="task_status",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "task_id": "The task ID to process",
        "task_type": 'The task type (either "restore" or "browse")',
    }
)
class TaskStatus(Resource):
    """The :class:`burpui.api.tasks.TaskStatus` resource allows you to
    follow a restore task.

    This resource is part of the :mod:`burpui.api.tasks` module.
    """

    if config["WITH_LIMIT"]:
        try:
            from ..ext.limit import limiter

            decorators = [limiter.exempt]
        except ImportError:
            pass

    @ns.doc(
        responses={
            200: "Success",
            500: "Task failed",
        },
    )
    def get(self, task_type, task_id, server=None):
        """Returns the state of the given task"""
        if task_type not in task_types:
            self.abort(400)
        task_obj, callback = task_types[task_type]
        task = task_obj.AsyncResult(task_id)
        if task.state == "FAILURE":
            if db:
                rec = Task.query.filter_by(uuid=task_id).first()
                if rec:
                    try:
                        db.session.delete(rec)
                        db.session.commit()
                    except:
                        db.session.rollback()
            task.revoke()
            err = str(task.result)
            self.abort(502, err)
        if task.state == "SUCCESS":
            if not task.result:
                self.abort(500, "The task did not return anything")
            server = task.result.get("server")
            return {
                "state": task.state,
                "location": url_for(callback, task_id=task_id, server=server),
            }
        return {"state": task.state}

    @ns.doc(
        responses={
            201: "Success",
            400: "Wrong request",
            403: "Permission denied",
        },
    )
    def delete(self, task_type, task_id, server=None):
        """Cancel a given task"""
        if task_type not in task_types:
            self.abort(400)
        task_obj, _ = task_types[task_type]
        task = task_obj.AsyncResult(task_id)
        user = task.result.get("user")
        dst_server = task.result.get("server")

        if (
            current_user.name != user or (dst_server and dst_server != server)
        ) and not current_user.acl.is_admin():
            self.abort(403, "Unauthorized access")

        # do not remove the task from db yet since we may need to remove
        # some temporary files afterward. The "cleanup_restore" task will take
        # care of this
        task.revoke()
        return "", 201


@ns.route("/get/<task_id>", "/<server>/get/<task_id>", endpoint="task_get_file")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "task_id": "The task ID to process",
    }
)
class TaskGetFile(Resource):
    """The :class:`burpui.api.tasks.TaskGetFile` resource allows you to
    retrieve the archive generated by the given task.

    This resource is part of the :mod:`burpui.api.tasks` module.
    """

    @ns.doc(
        responses={
            200: "Success",
            400: "Incomplete task",
            403: "Insufficient permissions",
            500: "Task failed",
        },
    )
    def get(self, task_id, server=None):
        """Returns the generated archive"""
        task = perform_restore.AsyncResult(task_id)
        bui.audit.logger.debug(f"downloading file for task {task_id}")
        if task.state != "SUCCESS":
            bui.audit.logger.info(f"unable to complete task: {task.state}")
            if task.state == "FAILURE":
                err = task.result.get("error")
                if err != "encrypted" and not task.result.get("admin"):
                    err = (
                        "An error occurred while performing the "
                        "restoration. Please contact your administrator "
                        "for further details"
                    )
                self.abort(500, "Unsuccessful task:\n{}".format(err))
            self.abort(400, "Task not processed yet: {}".format(task.state))

        path = task.result.get("path")
        user = task.result.get("user")
        dst_server = task.result.get("server")
        filename = task.result.get("filename")

        bui.audit.logger.info(
            f"restored file {filename} ({path}) for user {user}", server=dst_server
        )

        if (
            current_user.name != user or (dst_server and dst_server != server)
        ) and not current_user.acl.is_admin():
            bui.audit.logger.error("cannot send file: unauthorized access")
            self.abort(403, "Unauthorized access")

        if db:
            rec = Task.query.filter_by(uuid=task_id).first()
            if rec:
                try:
                    db.session.delete(rec)
                    db.session.commit()
                except:
                    db.session.rollback()
        task.revoke()

        if dst_server:
            return self.stream_file(path, filename, dst_server)

        try:
            # Trick to delete the file while sending it to the client.
            # First, we open the file in reading mode so that a file handler
            # is open on the file. Then we delete it as soon as the request
            # ended. Because the fh is open, the file will be actually removed
            # when the transfer is done and the send_file method has closed
            # the fh. Only tested on Linux systems.
            fh = open(path, "rb")

            @after_this_request
            def remove_file(response):
                """Callback function to run after the client has handled
                the request to remove temporary files.
                """
                os.remove(path)
                return response

            resp = send_file(
                fh,
                as_attachment=True,
                attachment_filename=filename,
                mimetype="application/zip",
            )
            resp.set_cookie("fileDownload", "true")
        except Exception as e:
            bui.client.logger.error(str(e))
            self.abort(500, str(e))

        return resp

    def stream_file(self, path, filename, server):
        socket = bui.client.get_file(path, server)
        if not socket:
            self.abort(500)
        lengthbuf = socket.recv(8)
        (length,) = struct.unpack("!Q", lengthbuf)

        def stream(sock, size):
            """The restoration took place on another server so we need
            to stream the file that is not present on the current
            machine.
            """
            bsize = 1024
            received = 0
            if size < bsize:
                bsize = size
            while received < size:
                buf = b""
                r, _, _ = select.select([sock], [], [], 5)
                if not r:
                    self.abort(500, "Socket timed-out")
                buf += sock.recv(bsize)
                if not buf:
                    continue
                received += len(buf)
                self.logger.debug("{}/{}".format(received, size))
                yield buf
            sock.sendall(struct.pack("!Q", 2))
            sock.sendall(b"RE")
            sock.close()

        headers = Headers()
        headers.add("Content-Disposition", "attachement", filename=filename)
        headers["Content-Length"] = length

        resp = Response(
            stream(socket, length),
            mimetype="application/zip",
            headers=headers,
            direct_passthrough=True,
        )
        resp.set_cookie("fileDownload", "true")
        resp.set_etag(
            "flask-%s-%s-%s"
            % (time(), length, adler32(filename.encode("utf-8")) & 0xFFFFFFFF)
        )

        return resp


@ns.route(
    "/archive/<name>/<int:backup>",
    "/<server>/archive/<name>/<int:backup>",
    endpoint="task_restore",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "name": "Client name",
        "backup": "Backup number",
    },
)
class TaskRestore(Resource):
    """The :class:`burpui.api.tasks.TaskRestore` resource allows you to
    perform a file restoration.

    This resource is part of the :mod:`burpui.api.tasks` module.

    The following parameters are supported:
    - ``list``: list of files/directories to restore
    - ``strip``: number of elements to strip in the path
    - ``format``: returning archive format
    - ``pass``: password to use for encrypted backups
    """

    parser = ns.parser()
    parser.add_argument(
        "pass", help="Password to use for encrypted backups", nullable=True
    )
    parser.add_argument(
        "format",
        required=False,
        help="Returning archive format",
        choices=("zip", "tar.gz", "tar.bz2"),
        default="zip",
        nullable=True,
    )
    parser.add_argument(
        "strip",
        type=int,
        help="Number of elements to strip in the path",
        default=0,
        nullable=True,
    )
    parser.add_argument(
        "list",
        required=True,
        help="List of files/directories to restore",
        nullable=False,
    )
    parser.add_argument(
        "timeout",
        type=int,
        required=False,
        help="Maximum task duration after you consider it stalled (in minutes)",
        default=60,
        nullable=True,
    )

    @ns.expect(parser, validate=True)
    @ns.doc(
        responses={
            202: "Accepted",
            400: "Missing parameter",
            403: "Insufficient permissions",
        },
    )
    def post(self, server=None, name=None, backup=None):
        """Performs an online restoration

        **POST** method provided by the webservice.

        This method returns a :mod:`flask.Response` object.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int
        """
        args = self.parser.parse_args()
        files = args["list"]
        strip = args["strip"]
        fmt = args["format"] or "zip"
        passwd = args["pass"]
        timeout = args["timeout"]
        args_log = args.copy()
        # don't leak secrets in logs
        del args_log["pass"]
        bui.audit.logger.info(
            f"requested restoration of backup n°{backup} for {name}: {args_log}",
            server=server,
        )
        room = None
        if WS_AVAILABLE:
            room = request.sid
        if not files or not name or not backup:
            self.abort(400, "missing arguments")
        # Manage ACL
        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_client_rw(name, server)
        ):
            self.abort(
                403, "You are not allowed to perform a restoration for this client"
            )
        task = perform_restore.apply_async(
            args=[
                name,
                backup,
                files,
                strip,
                fmt,
                passwd,
                server,
                current_user.name,
                not current_user.is_anonymous and current_user.acl.is_admin(),
                room,
            ]
        )
        if db:
            db_task = Task(
                task.id,
                "perform_restore",
                current_user.name,
                timedelta(minutes=timeout),
            )
            try:
                db.session.add(db_task)
                db.session.commit()
            except:
                db.session.rollback()
        return {"id": task.id, "name": "perform_restore"}, 202


@ns.route(
    "/config/<client>",
    "/config/<client>/<path:conf>",
    "/<server>/config/<client>",
    "/<server>/config/<client>/<path:conf>",
    endpoint="task_delete_client",
    methods=["DELETE"],
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "client": "Client name",
        "conf": "Path of the configuration file",
    },
)
class ClientSettings(Resource):
    parser_delete = ns.parser()
    parser_delete.add_argument(
        "revoke",
        type=inputs.boolean,
        help="Whether to revoke the certificate or not",
        default=False,
        nullable=True,
    )
    parser_delete.add_argument(
        "delcert",
        type=inputs.boolean,
        help="Whether to delete the certificate or not",
        default=False,
        nullable=True,
    )
    parser_delete.add_argument(
        "keepconf",
        type=inputs.boolean,
        help="Whether to keep the conf or not",
        default=False,
        nullable=True,
    )
    parser_delete.add_argument(
        "template",
        type=inputs.boolean,
        help="Whether we work on a template or not",
        default=False,
        nullable=True,
    )
    parser_delete.add_argument(
        "delete",
        type=inputs.boolean,
        help="Whether we should remove the data as well or not",
        default=False,
        nullable=True,
    )

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(
        message=_("Sorry, you don't have rights to access the setting panel")
    )
    @ns.expect(parser_delete)
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
            409: "Conflict",
            500: "Internal failure",
        }
    )
    def delete(self, server=None, client=None, conf=None):
        """Deletes a given client"""
        if (
            not current_user.is_anonymous
            and current_user.acl.is_moderator()
            and not current_user.acl.is_server_rw(server)
        ):
            self.abort(403, "You don't have rights on this server")

        if bui.client.is_backup_running(client, server):
            self.abort(
                409,
                "There is currently a backup running for this client hence "
                "we cannot delete it for now. Please try again later",
            )

        args = self.parser_delete.parse_args()
        delcert = args.get("delcert", False)
        revoke = args.get("revoke", False)
        keepconf = args.get("keepconf", False)
        template = args.get("template", False)
        delete = args.get("delete", False)

        task = delete_client.apply_async(
            args=[
                client,
                keepconf,
                delcert,
                revoke,
                template,
                delete,
                server,
                current_user.name,
            ]
        )

        if db:
            db_task = Task(
                task.id, "delete_client", current_user.name, timedelta(minutes=60)
            )
            try:
                db.session.add(db_task)
                db.session.commit()
            except:
                db.session.rollback()
        return {"id": task.id, "name": "delete_client"}, 202


@ns.route(
    "/completed/config/<task_id>",
    "/completed/<server>/config/<task_id>",
    endpoint="task_deleted_client",
)
@ns.doc(
    params={
        "task_id": "The task ID to process",
    }
)
class TaskDeletedClient(Resource):
    """The :class:`burpui.api.tasks.TaskDeletedClient` resource allows you to
    retrieve the result of the delete_client task.

    This resource is part of the :mod:`burpui.api.tasks` module.
    """

    @ns.doc(
        responses={
            400: "Incomplete task",
            403: "Insufficient permissions",
            500: "Task failed",
        },
    )
    def get(self, task_id, server=None):
        """Returns the task result"""
        task = delete_client.AsyncResult(task_id)
        if task.state != "SUCCESS":
            if task.state == "FAILURE":
                self.abort(
                    500, "Unsuccessful task: {}".format(task.result.get("error"))
                )
            self.abort(400, "Task not processed yet: {}".format(task.state))

        tres = task.result

        user = tres.get("user")
        dst_server = tres.get("server")
        resp = tres.get("result")
        kwargs = tres.get("kwargs")

        client = tres.get("client")
        delcert = kwargs.get("delcert")
        revoke = kwargs.get("revoke")
        keepconf = kwargs.get("keepconf")
        delete = kwargs.get("delete")
        template = kwargs.get("template")

        if current_user.name != user or (dst_server and dst_server != server):
            self.abort(403, "Unauthorized access")

        task.revoke()

        if not keepconf:
            # clear the cache when we remove a client
            cache.clear()
            # clear client-side cache through the _extra META variable
            try:
                _extra = session.get("_extra", g.now)
                _extra = int(_extra)
            except ValueError:
                _extra = 0
            session["_extra"] = "{}".format(_extra + 1)
            if bui.config["WITH_CELERY"]:
                force_scheduling_now()

        bui.audit.logger.info(
            f"deleted client configuration {client}, delete certificate: {delcert}, "
            f"revoke certificate: {revoke}, keep a backup of the configuration: "
            f"{keepconf}, delete data: {delete}, is template: {template}",
            server=server,
        )
        return resp


#        if not keepconf:
#        parser = bui.client.get_parser(agent=server)
#
#        bui.audit.logger.info(f'deleted client configuration {client} ({conf}), delete certificate: {delcert}, revoke certificate: {revoke}, keep a backup of the configuration: {keepconf}, delete data: {delete}', server=server)
#        return parser.remove_client(client, keepconf, delcert, revoke, template, delete), 200


@ns.route(
    "/running",
    "/<server>/running",
    "/running/<client>",
    "/<server>/running/<client>",
    endpoint="task_running_clients",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "client": "Client name",
    },
)
class TaskRunningClients(RunningClients):
    """The :class:`burpui.api.tasks.TaskRunningClients` resource allows you
    to retrieve a list of clients that are currently running a backup.

    This resource is part of the :mod:`burpui.api.tasks` module.

    This resource is backed by a periodic task. If the periodic task fail or is
    not running, we fallback to the "synchronous" API call.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.

    .. seealso:: :class:`burpui.api.clients.RunningClients`
    """

    def get(self, client=None, server=None):
        """Returns a list of clients currently running a backup

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [ 'client1', 'client2' ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to see.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :param client: Ask a specific client in order to know if it is running a backup
        :type client: str

        :returns: The *JSON* described above.
        """
        server = server or self.parser.parse_args()["serverName"]
        res = cache.cache.get("backup_running_result")
        if res is None:
            res = bui.client.is_one_backup_running(server)
        return self._running_clients(res, client, server)


@ns.route("/backup-running", "/<server>/backup-running", endpoint="task_running_backup")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    }
)
class TaskRunningBackup(RunningBackup):
    """The :class:`burpui.api.tasks.TaskRunningBackup` resource allows you to
    access the status of the server in order to know if there is a running
    backup currently.

    This resource is backed by a periodic task. If the periodic task fail or is
    not running, we fallback to the "synchronous" API call.

    This resource is part of the :mod:`burpui.api.tasks` module.
    """

    @ns.marshal_with(
        RunningBackup.running_fields,
        code=200,
        description="Success",
    )
    def get(self, server=None):
        """Tells if a backup is running right now

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
                "running": false
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :returns: The *JSON* described above.
        """
        res = cache.cache.get("backup_running_result")
        if res is None:
            res = bui.client.is_one_backup_running(server)
        return {"running": self._is_one_backup_running(res, server)}


@ns.route(
    "/history",
    "/history/<client>",
    "/<server>/history",
    "/<server>/history/<client>",
    endpoint="task_history",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "client": "Client name",
    },
)
class TaskHistory(History):
    """The :class:`burpui.api.misc.History` resource allows you to retrieve
    an history of the backups

    This resource is backed by a periodic task. If the periodic task fail or is
    not running, we redirect to the "synchronous" API call.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode and ``clientName`` is also allowed to filter
    by client.

    ::

        $('#calendar').fullCalendar({

            eventSources: [

                // your event source
                {
                    events: [ // put the array in the `events` property
                        {
                            title  : 'event1',
                            start  : '2010-01-01'
                        },
                        {
                            title  : 'event2',
                            start  : '2010-01-05',
                            end    : '2010-01-07'
                        },
                        {
                            title  : 'event3',
                            start  : '2010-01-09T12:30:00',
                        }
                    ],
                    color: 'black',     // an option!
                    textColor: 'yellow' // an option!
                }

                // any other event sources...

            ]

        });

    """

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_with(
        History.history_fields, code=200, description="Success", as_list=True
    )
    @ns.expect(History.parser)
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
        },
    )
    @browser_cache(1800)
    def get(self, client=None, server=None):
        """Returns a list of calendars describing the backups that have been
        completed so far

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                "color": "#7C6F44",
                "events": [
                  {
                    "backup": "0000001",
                    "end": "2015-01-25 13:32:04+01:00",
                    "name": "toto-test",
                    "start": "2015-01-25 13:32:00+01:00",
                    "title": "Client: toto-test, Backup n°0000001",
                    "url": "/client/toto-test"
                  }
                ],
                "name": "toto-test",
                "textColor": "white"
              }
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str
        :param client: Which client to collect data from
        :type client: str

        :returns: The *JSON* described above
        """
        self._check_acl(client, server)
        res = cache.cache.get("all_backups")
        if res is None:
            return self._get_backup_history(client, server)

        return self._get_backup_history(client, server, res)


@ns.route("/report", "/<server>/report", endpoint="task_clients_report")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class TaskClientsReport(ClientsReport):
    """The :class:`burpui.api.tasks.TaskClientsReport` resource allows you to
    access general reports about your clients.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_with(
        ClientsReport.report_fields,
        code=200,
        description="Success",
    )
    @ns.expect(ClientsReport.parser)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            500: "Internal failure",
        },
    )
    @browser_cache(1800)
    def get(self, server=None):
        """Returns a global report about all the clients of a given server

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "backups": [
                {
                  "name": "client1",
                  "number": 15
                },
                {
                  "name": "client2",
                  "number": 1
                }
              ],
              "clients": [
                {
                  "name": "client1",
                  "stats": {
                    "total": 296377,
                    "totsize": 57055793698,
                    "windows": "unknown"
                  }
                },
                {
                  "name": "client2",
                  "stats": {
                    "total": 3117,
                    "totsize": 5345361,
                    "windows": "true"
                  }
                }
              ]
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :returns: The *JSON* described above
        """
        server = server or self.parser.parse_args()["serverName"]
        self._check_acl(server)
        res = cache.cache.get("all_clients_reports")
        if res is None:
            return self._get_clients_reports(server=server)

        return self._get_clients_reports(res, server)


@ns.route(
    "/browseall/<name>/<int:backup>",
    "/<server>/browsall/<name>/<int:backup>",
    endpoint="task_client_tree_all",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in" + " multi-agent mode",
        "name": "Client name",
        "backup": "Backup number",
    },
)
class TaskClientTreeAll(Resource):
    """The :class:`burpui.api.tasks.TaskClientTreeAll` resource allows you to
    retrieve a list of all the files in a given backup through the celery
    worker.

    This resource is part of the :mod:`burpui.api.client` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )

    @ns.expect(parser)
    @ns.doc(
        responses={
            202: "Accepted",
            405: "Method not allowed",
            403: "Insufficient permissions",
            500: "Internal failure",
        },
    )
    def post(self, server=None, name=None, backup=None):
        """Launch the tasks that will gather all nodes of a given backup

        **POST** method provided by the webservice.

        This method returns a :mod:`flask.Response` object.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int
        """
        args = self.parser.parse_args()
        server = server or args.get("serverName")

        if not bui.client.get_attr("batch_list_supported", False, server):
            self.abort(405, "Sorry, the requested backend does not support this method")

        # Manage ACL
        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_client_allowed(name, server)
        ):
            self.abort(403, "Sorry, you are not allowed to view this client")

        task = load_all_tree.apply_async(args=[name, backup, server, current_user.name])
        return {"id": task.id, "name": "load_all_tree"}, 202


@ns.route(
    "/get-browse/<task_id>",
    "/<server>/get-browse/<task_id>",
    endpoint="task_do_browse_all",
)
@ns.doc(
    params={
        "task_id": "The task ID to process",
    }
)
class TaskDoBrowseAll(Resource):
    """The :class:`burpui.api.tasks.TaskDoBrowseAll` resource allows you to
    retrieve the tree generated by the given task.

    This resource is part of the :mod:`burpui.api.tasks` module.
    """

    @ns.marshal_list_with(node_fields, code=200, description="Success")
    @ns.doc(
        responses={
            400: "Incomplete task",
            403: "Insufficient permissions",
            500: "Task failed",
        },
    )
    def get(self, task_id, server=None):
        """Returns the generated archive"""
        task = load_all_tree.AsyncResult(task_id)
        if task.state != "SUCCESS":
            if task.state == "FAILURE":
                self.abort(
                    500, "Unsuccessful task: {}".format(task.result.get("error"))
                )
            self.abort(400, "Task not processed yet: {}".format(task.state))

        user = task.result.get("user")
        dst_server = task.result.get("server")
        resp = task.result.get("tree")

        if current_user.name != user or (dst_server and dst_server != server):
            self.abort(403, "Unauthorized access")

        task.revoke()

        return resp



"""user prefs

Revision ID: 56de018f4d88
Revises: fc07e3fa0086
Create Date: 2017-02-12 14:51:38.147422

"""

# revision identifiers, used by Alembic.
revision = "56de018f4d88"
down_revision = "fc07e3fa0086"

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "pref",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user", sa.String(length=256), nullable=True),
        sa.Column("key", sa.String(length=256), nullable=False),
        sa.Column("value", sa.String(length=256), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("pref")
    ### end Alembic commands ###



# -*- coding: utf8 -*-
"""

.. module:: burpui.extensions
    :platform: Unix
    :synopsis: Burp-UI extensions module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import re
import warnings

from ._compat import to_unicode


def parse_db_setting(string):
    parts = re.search(
        r"(?:(?P<backend>\w+)(?:\+(?P<driver>\w+))?://)?"
        r"(?:(?P<user>\w+)(?::?(?P<pass>.+))?@)?"
        r"(?P<host>[\w_.-]+):?(?P<port>\d+)?(?:/(?P<db>\w+))?",
        string,
    )
    if not parts:  # pragma: no cover
        raise ValueError('Unable to parse the db: "{}"'.format(string))
    back = parts.group("backend") or ""
    user = parts.group("user") or None
    pwd = parts.group("pass") or None
    host = parts.group("host") or ""
    port = parts.group("port") or ""
    db = parts.group("db") or ""
    return (back, user, pwd, host, port, db)


def get_redis_server(myapp):
    host = "localhost"
    port = 6379
    pwd = None
    if myapp.redis and myapp.redis.lower() != "none":
        try:
            back, user, pwd, host, port, db = parse_db_setting(myapp.redis)
            host = host or "localhost"
            try:
                port = int(port)
            except (ValueError, IndexError):
                port = 6379
        except ValueError:  # pragma: no cover
            pass
    return host, port, pwd


def create_db(myapp, cli=False, unittest=False, create=True, celery_worker=False):
    """Create the SQLAlchemy instance if possible

    :param myapp: Application context
    :type myapp: :class:`burpui.engines.server.BUIServer`
    """
    if myapp.config["WITH_SQL"]:
        try:
            from .ext.sql import db
            from sqlalchemy.exc import OperationalError
            from sqlalchemy_utils.functions import database_exists
            from .models import lazy_loading

            lazy_loading()
            myapp.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
            if (
                not database_exists(myapp.config["SQLALCHEMY_DATABASE_URI"])
                and not cli
                and not unittest
                and not celery_worker
            ):
                if create:  # pragma: no cover
                    import subprocess

                    local = os.path.join(os.getcwd(), "..", "tools", "bui-manage")
                    buimanage = local if os.path.exists(local) else "bui-manage"
                    cmd = [
                        buimanage,
                        "-c",
                        myapp.config["CFG"],
                        "-l",
                        os.devnull,
                        "db",
                        "upgrade",
                    ]
                    upgd = subprocess.Popen(
                        cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                    )
                    (out, _) = upgd.communicate()
                    if upgd.returncode != 0:
                        myapp.logger.error(
                            "Disabling SQL support because "
                            "something went wrong while setting up the "
                            "database:\n{}".format(out)
                        )
                        myapp.config["WITH_SQL"] = False
                        return None
                    return create_db(myapp, cli, unittest, False)
                else:  # pragma: no cover
                    myapp.logger.error("Database not found, disabling SQL support")
                    myapp.config["WITH_SQL"] = False
                    return None

            back = parse_db_setting(myapp.config["SQLALCHEMY_DATABASE_URI"])[0]

            if "mysql" in back:  # pragma: no cover
                # optimize SQL pools for MySQL driver
                myapp.config["SQLALCHEMY_POOL_SIZE"] = 20
                myapp.config["SQLALCHEMY_POOL_RECYCLE"] = 600

            db.init_app(myapp)
            if not cli and not unittest and not celery_worker:  # pragma: no cover
                with myapp.app_context():
                    try:
                        import subprocess

                        # get current head using alembic/FLask-Migrate
                        local = os.path.join(os.getcwd(), "tools", "bui-manage")
                        buimanage = local if os.path.exists(local) else "bui-manage"
                        cmd = [
                            buimanage,
                            "-c",
                            myapp.config["CFG"],
                            "-l",
                            os.devnull,
                            "db",
                            "current",
                        ]
                        rev = subprocess.Popen(
                            cmd, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
                        )
                        (out, _) = rev.communicate()
                        if rev.returncode != 0:
                            raise Exception(
                                "something went wrong while setting up the "
                                "database:\n{}".format(out)
                            )

                        current = to_unicode(out)

                        # now we compare the revision numbers
                        if "head" not in current:
                            myapp.logger.critical(
                                "Your database seems out of sync, "
                                "you may want to run 'bui-manage db "
                                "upgrade'."
                            )
                            myapp.logger.critical("Disabling SQL support for now.")
                            myapp.config["WITH_SQL"] = False
                            return None

                    except (OperationalError, Exception) as exp:
                        err = str(exp)
                        if "no such table" in err:
                            myapp.logger.critical(
                                "Your database seems out of sync, you may want "
                                "to run 'bui-manage db upgrade'."
                            )
                        else:
                            myapp.logger.critical(
                                "Something seems to be wrong with your setup: "
                                "{}".format(err)
                            )

                        myapp.logger.critical("Disabling SQL support for now.")
                        myapp.config["WITH_SQL"] = False
                        return None

            # If we are here, it means everything is alright
            return db

        except ImportError:  # pragma: no cover
            myapp.logger.critical(
                "Unable to load requirements, you may want to run 'pip "
                'install "burp-ui[sql]"\'.\nDisabling SQL support for now.'
            )
            myapp.config["WITH_SQL"] = False
        except OperationalError as exp:  # pragma: no cover
            myapp.logger.critical(
                "unable to contact database: {}\nDisabling SQL " "support.".format(exp)
            )
            myapp.config["WITH_SQL"] = False

    return None


def create_websocket(myapp, websocket_server=False, celery_worker=False, cli=False):
    """Create the websocket server if possible

    :param myapp: Application context
    :type myapp: :class:`burpui.engines.server.BUIServer`
    """
    if cli and not websocket_server:
        return False
    broker = myapp.ws_broker
    if broker is not False:
        if not broker or broker is True:
            broker = "redis"
    if broker and broker.lower() != "none":
        host, oport, pwd = get_redis_server(myapp)
        odb = 4
        if broker.lower() not in ["default", "redis"]:
            try:
                (_, _, pwd, host, port, db) = parse_db_setting(myapp.use_celery)
                if not port:
                    port = oport
                if not db:
                    db = odb
                else:
                    try:
                        db = int(db)
                    except ValueError:
                        db = odb
            except ValueError:
                pass
        else:
            port = oport
            db = odb
        if pwd:
            redis_url = "redis://:{}@{}:{}/{}".format(pwd, host, port, db)
        else:
            redis_url = "redis://{}:{}/{}".format(host, port, db)
        myapp.config["WS_MESSAGE_QUEUE"] = redis_url
    myapp.config["WS_MANAGE_SESSION"] = not myapp.config.get("WITH_SRV_SESSION", False)
    if os.getenv("BUI_MODE", "") == "celery":
        myapp.config["WS_ASYNC_MODE"] = "threading"

    if celery_worker:
        return False

    # if you are not a celery worker, we can patch the flask server
    try:
        from .ext.ws import socketio

        socketio.init_app(
            myapp,
            message_queue=myapp.config.get("WS_MESSAGE_QUEUE"),
            manage_session=myapp.config.get("WS_MANAGE_SESSION", False),
            async_mode=myapp.config.get("WS_ASYNC_MODE", "gevent"),
        )
        myapp.config["WS_AVAILABLE"] = True
    except ImportError:
        myapp.config["WS_AVAILABLE"] = False

    # Now load the namespaces
    if myapp.config["WITH_WS"] or websocket_server:
        from .ws.namespace import BUINamespace

        socketio.on_namespace(BUINamespace("/ws"))
        return True

    return False


def create_celery(myapp, warn=True):
    """Create the Celery app if possible

    :param myapp: Application context
    :type myapp: :class:`burpui.engines.server.BUIServer`
    """
    if myapp.config["WITH_CELERY"]:  # pragma: no cover
        from .exceptions import BUIserverException

        host, oport, pwd = get_redis_server(myapp)
        odb = 2
        if isinstance(myapp.use_celery, str):
            try:
                (_, _, pwd, host, port, db) = parse_db_setting(myapp.use_celery)
                if not port:
                    port = oport
                if not db:
                    db = odb
                else:
                    try:
                        db = int(db)
                    except ValueError:
                        db = odb
            except ValueError:
                pass
        else:
            db = odb
            port = oport
        if pwd:
            redis_url = "redis://:{}@{}:{}/{}".format(pwd, host, port, db)
        else:
            redis_url = "redis://{}:{}/{}".format(host, port, db)
        myapp.config["CELERY_BROKER_URL"] = myapp.config["BROKER_URL"] = redis_url
        myapp.config["CELERY_RESULT_BACKEND"] = redis_url

        from .ext.tasks import celery

        celery.conf.update(myapp.config)

        if not hasattr(celery, "flask_app"):
            celery.flask_app = myapp

        TaskBase = celery.Task

        class ContextTask(TaskBase):
            abstract = True

            def __call__(self, *args, **kwargs):
                with myapp.app_context():
                    try:
                        return TaskBase.__call__(self, *args, **kwargs)
                    except BUIserverException:
                        # ignore unhandled exceptions in the celery worker
                        pass

        celery.Task = ContextTask

        # may fail in case redis is not running (this can happen while running
        # the bui-manage script)
        try:
            if os.getenv("BUI_MODE", "") == "celery":
                from .tasks import force_scheduling_now

                force_scheduling_now()
        except:  # pragma: no cover
            pass

        return celery

    if warn:  # pragma: no cover
        message = (
            "Something went wrong while initializing celery worker.\n"
            "Maybe it is not enabled in your conf "
            "({}).".format(myapp.config["CFG"])
        )
        warnings.warn(message, RuntimeWarning)

    return None



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.acl.interface
    :platform: Unix
    :synopsis: Burp-UI ACL interface.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from abc import ABCMeta, abstractmethod, abstractproperty

import logging


class BUIaclLoader(object, metaclass=ABCMeta):
    """The :class:`burpui.misc.acl.interface.BUIaclLoader` class is used to
    load the actual ACL backend"""

    logger = logging.getLogger("burp-ui")
    priority = 0

    add_grant = False
    del_grant = False
    mod_grant = False

    add_group = False
    del_group = False
    mod_group = False

    add_group_member = False
    del_group_member = False

    add_moderator = False
    del_moderator = False
    mod_moderator = False

    add_admin = False
    del_admin = False

    moderator = None
    moderators = []

    admins = []

    def __init__(self, app=None):
        """:func:`burpui.misc.acl.interface.BUIaclLoader.__init__` instanciate
        the loader.

        :param app: Application context
        :type app: :class:`burpui.engines.server.BUIServer`
        """
        self.app = None
        if app:  # pragma: no cover
            self.init_app(app)

    def init_app(self, app):
        """Register the given app"""
        self.app = app

    @abstractmethod
    def reload(self):
        """Reload the backend"""
        return None  # pragma: no cover

    @abstractproperty
    @property
    def acl(self):
        """Property to retrieve the backend"""
        return None  # pragma: no cover

    @abstractproperty
    @property
    def grants(self):
        """Property to retrieve the list of grants"""
        return None  # pragma: no cover

    @abstractproperty
    @property
    def groups(self):
        """Property to retrieve the list of groups with their members"""
        return None  # pragma: no cover


class BUIacl(object, metaclass=ABCMeta):
    """The :class:`burpui.misc.acl.interface.BUIacl` class represents the ACL
    engine.
    """

    def __init__(self, app=None):
        """:func:`burpui.misc.acl.interface.BUIacl.__init__` instanciate
        the ACL.

        :param app: Application context
        :type app: :class:`burpui.engines.server.BUIServer`
        """
        self.app = None
        if app:  # pragma: no cover
            self.init_app(app)

    def init_app(self, app):
        """Register the given app"""
        self.app = app

    @abstractmethod
    def is_admin(self, username=None):
        """:func:`burpui.misc.acl.interface.BUIacl.is_admin` is used to know if
        a user has administrator rights.

        :param username: Username to check
        :type username: str

        :returns: True if the user has admin rights, otherwise False
        :rtype: bool
        """
        return False  # pragma: no cover

    @abstractmethod
    def is_moderator(self, username=None):
        """:func:`burpui.misc.acl.interface.BUIacl.is_moderator` is used to
        know if a user has moderator rights.

        :param username: Username to check
        :type username: str

        :returns: True if the user has moderator rights, otherwise False
        :rtype: bool
        """
        return False  # pragma: no cover

    @abstractmethod
    def is_client_rw(self, username=None, client=None, server=None):
        """:func:`burpui.misc.acl.interface.BUIacl.is_client_rw` tells us
        if a given user has access to a given client on a given server in RW
        mode.

        :param username: Username to check
        :type username: str

        :param client: Client to check
        :type client: str

        :param server: Server to check
        :type server: str

        :returns: True if username is granted, otherwise False
        :rtype: bool
        """
        return False  # pragma: no cover

    @abstractmethod
    def is_client_allowed(self, username=None, client=None, server=None):
        """:func:`burpui.misc.acl.interface.BUIacl.is_client_allowed` tells us
        if a given user has access to a given client on a given server.

        :param username: Username to check
        :type username: str

        :param client: Client to check
        :type client: str

        :param server: Server to check
        :type server: str

        :returns: True if username is granted, otherwise False
        :rtype: bool
        """
        return False  # pragma: no cover

    @abstractmethod
    def is_server_rw(self, username=None, server=None):
        """:func:`burpui.misc.acl.interface.BUIacl.is_server_rw` tells us
        if a given user has access to a given server in RW mode.

        :param username: Username to check
        :type username: str

        :param server: Server to check
        :type server: str

        :returns: True if username is granted, otherwise False
        :rtype: bool
        """
        return False  # pragma: no cover

    @abstractmethod
    def is_server_allowed(self, username=None, server=None):
        """:func:`burpui.misc.acl.interface.BUIacl.is_server_allowed` tells us
        if a given user has access to a given server.

        :param username: Username to check
        :type username: str

        :param server: Server to check
        :type server: str

        :returns: True if username is granted, otherwise False
        :rtype: bool
        """
        return False  # pragma: no cover



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.clients
    :platform: Unix
    :synopsis: Burp-UI clients api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from . import api, cache_key, force_refresh
from ..engines.server import BUIServer  # noqa
from .custom import fields, Resource
from .client import ClientLabels
from ..ext.cache import cache
from ..exceptions import BUIserverException
from ..decorators import browser_cache
from ..filter import mask

from flask import current_app, g
from flask_login import current_user

bui = current_app  # type: BUIServer
ns = api.namespace("clients", "Clients methods")


@ns.route(
    "/running",
    "/<server>/running",
    "/running/<client>",
    "/<server>/running/<client>",
    endpoint="running_clients",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "client": "Client name",
    },
)
class RunningClients(Resource):
    """The :class:`burpui.api.clients.RunningClients` resource allows you to
    retrieve a list of clients that are currently running a backup.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )

    @ns.expect(parser)
    def get(self, client=None, server=None):
        """Returns a list of clients currently running a backup

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              'client1',
              'client2'
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to see.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :param client: Ask a specific client in order to know if it is running a backup
        :type client: str

        :returns: The *JSON* described above.
        """
        server = server or self.parser.parse_args()["serverName"]
        return self._running_clients(None, client, server)

    def _running_clients(self, res, client, server):
        if client:
            if mask.has_filters(current_user):
                if not mask.is_client_allowed(current_user, client):
                    return []

            if bui.client.is_backup_running(client, server):
                return [client]
            else:
                return []

        running = res or bui.client.is_one_backup_running(server)
        # Manage ACL
        if mask.has_filters(current_user):
            if isinstance(running, dict):
                ret = {}

                def __extract_running_clients(serv):
                    try:
                        clients = [
                            x["name"]
                            for x in bui.client.get_all_clients(
                                serv, last_attempt=False
                            )
                        ]
                    except BUIserverException:
                        clients = []
                    allowed = [
                        x
                        for x in clients
                        if mask.is_client_allowed(current_user, x, serv)
                    ]
                    return [x for x in running[serv] if x in allowed]

                if server:
                    return __extract_running_clients(server)

                for serv in bui.client.servers:
                    ret[serv] = __extract_running_clients(serv)
                return ret
            else:
                try:
                    clients = [
                        x["name"]
                        for x in bui.client.get_all_clients(server, last_attempt=False)
                    ]
                except BUIserverException:
                    clients = []
                allowed = [
                    x
                    for x in clients
                    if mask.is_client_allowed(current_user, x, server)
                ]
                running = [x for x in running if x in allowed]
        elif server and isinstance(running, dict):
            return running.get(server, [])
        return running


@ns.route("/backup-running", "/<server>/backup-running", endpoint="running_backup")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    }
)
class RunningBackup(Resource):
    """The :class:`burpui.api.clients.RunningBackup` resource allows you to
    access the status of the server in order to know if there is a running
    backup currently.

    This resource is part of the :mod:`burpui.api.clients` module.
    """

    running_fields = ns.model(
        "Running",
        {
            "running": fields.Boolean(
                required=True, description="Is there a backup running right now"
            ),
        },
    )

    @ns.marshal_with(running_fields, code=200, description="Success")
    def get(self, server=None):
        """Tells if a backup is running right now

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
                "running": false
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above.
        """
        return {
            "running": self._is_one_backup_running(
                bui.client.is_one_backup_running(server), server
            )
        }

    def _is_one_backup_running(self, res, server):
        """Check if a backup is running"""
        # Manage ACL
        if mask.has_filters(current_user):
            if isinstance(res, dict):
                new = {}
                for serv in bui.client.servers:
                    try:
                        clients = [
                            x["name"]
                            for x in bui.client.get_all_clients(
                                serv, last_attempt=False
                            )
                        ]
                    except BUIserverException:
                        clients = []
                    allowed = [
                        x
                        for x in clients
                        if mask.is_client_allowed(current_user, x, serv)
                    ]
                    new[serv] = [x for x in res[serv] if x in allowed]
                res = new
            else:
                try:
                    clients = [
                        x["name"]
                        for x in bui.client.get_all_clients(server, last_attempt=False)
                    ]
                except BUIserverException:
                    clients = []
                allowed = [
                    x
                    for x in clients
                    if mask.is_client_allowed(current_user, x, server)
                ]
                res = [x for x in res if x in allowed]
        running = False
        if isinstance(res, dict):
            for (_, run) in res.items():
                running = running or (len(run) > 0)
                if running:
                    break
        else:
            running = len(res) > 0

        return running


@ns.route("/report", "/<server>/report", endpoint="clients_report")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class ClientsReport(Resource):
    """The :class:`burpui.api.clients.ClientsReport` resource allows you to
    access general reports about your clients.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    parser.add_argument(
        "limit", type=int, default=8, help="Number of elements to return"
    )
    parser.add_argument(
        "aggregation",
        help="What aggregation to operate",
        default="number",
        choices=("number", "files", "size", "none"),
    )

    translation = {
        "number": "number",
        "files": "total",
        "size": "totsize",
    }

    stats_fields = ns.model(
        "ClientsStats",
        {
            "total": fields.Integer(
                required=True, description="Number of files", default=0
            ),
            "totsize": fields.Integer(
                required=True,
                description="Total size occupied by all the backups of this client",
                default=0,
            ),
            "os": fields.String(
                required=True, description="OS of the client", default="unknown"
            ),
        },
    )
    client_fields = ns.model(
        "ClientsReport",
        {
            "name": fields.String(required=True, description="Client name"),
            "stats": fields.Nested(stats_fields, required=True),
        },
    )
    backup_fields = ns.model(
        "ClientsBackup",
        {
            "name": fields.String(required=True, description="Client name"),
            "number": fields.Integer(
                required=True, description="Number of backups on this client", default=0
            ),
        },
    )
    report_fields = ns.model(
        "Report",
        {
            "backups": fields.Nested(backup_fields, as_list=True, required=True),
            "clients": fields.Nested(client_fields, as_list=True, required=True),
        },
    )

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_with(report_fields, code=200, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            500: "Internal failure",
        },
    )
    @browser_cache(1800)
    def get(self, server=None):
        """Returns a global report about all the clients of a given server

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "backups": [
                {
                  "name": "client1",
                  "number": 15
                },
                {
                  "name": "client2",
                  "number": 1
                }
              ],
              "clients": [
                {
                  "name": "client1",
                  "stats": {
                    "total": 296377,
                    "totsize": 57055793698,
                    "os": "unknown"
                  }
                },
                {
                  "name": "client2",
                  "stats": {
                    "total": 3117,
                    "totsize": 5345361,
                    "os": "windows"
                  }
                }
              ]
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above
        """

        server = server or self.parser.parse_args()["serverName"]
        self._check_acl(server)
        return self._get_clients_reports(server=server)

    def _check_acl(self, server):
        # Manage ACL
        if (
            not bui.config["STANDALONE"]
            and not current_user.is_anonymous
            and (
                not current_user.acl.is_admin()
                and not current_user.acl.is_server_allowed(server)
            )
        ):
            self.abort(403, "Sorry, you don't have any rights on this server")

    def _get_clients_reports(self, res=None, server=None):
        args = self.parser.parse_args()
        limit = args["limit"]
        aggregation = self.translation.get(args["aggregation"], "number")
        ret = self._parse_clients_reports(res, server)
        backups = backups_orig = ret.get("backups", [])
        clients = clients_orig = ret.get("clients", [])
        aggregate = False
        if limit > 1:
            limit -= 1
            aggregate = True
        if aggregation == "none":
            aggregate = False
            limit = 0
        # limit the number of elements to return so the graphs stay readable
        if len(backups) > limit and limit > 0:
            if aggregation == "number":
                backups = (
                    sorted(backups, key=lambda x: x.get("number"), reverse=True)
                )[:limit]
            else:
                clients = (
                    sorted(
                        clients,
                        key=lambda x: x.get("stats", {}).get(aggregation),
                        reverse=True,
                    )
                )[:limit]
        else:
            aggregate = False
        if aggregation == "number":
            clients_name = [x.get("name") for x in backups]
            ret["backups"] = backups
            ret["clients"] = [x for x in clients_orig if x.get("name") in clients_name]
        else:
            clients_name = [x.get("name") for x in clients]
            ret["clients"] = clients
            ret["backups"] = [x for x in backups_orig if x.get("name") in clients_name]
        if aggregate:
            backups = {"name": "others", "number": 0}
            for client in backups_orig:
                if client.get("name") not in clients_name:
                    backups["number"] += client.get("number", 0)

            complement = {
                "name": "others",
                "stats": {"total": 0, "totsize": 0, "os": None},
            }
            # TODO: fix OS aggregation
            for client in clients_orig:
                if client.get("name") not in clients_name:
                    complement["stats"]["total"] += client.get("stats", {}).get(
                        "total", 0
                    )
                    complement["stats"]["totsize"] += client.get("stats", {}).get(
                        "totsize", 0
                    )
                    os = client.get("stats", {}).get("os", "unknown")
                    if not complement["stats"]["os"]:
                        complement["stats"]["os"] = os
                    elif os != complement["stats"]["os"]:
                        complement["stats"]["os"] = "unknown"

            ret["clients"].append(complement)
            ret["backups"].append(backups)

        return ret

    def _parse_clients_reports(self, res=None, server=None):
        if not res:
            try:
                clients = bui.client.get_all_clients(agent=server, last_attempt=False)
            except BUIserverException as e:
                self.abort(500, str(e))
            if mask.has_filters(current_user):
                clients = [
                    x
                    for x in clients
                    if mask.is_client_allowed(current_user, x["name"], server)
                ]
            return bui.client.get_clients_report(clients, server)
        if bui.config["STANDALONE"]:
            ret = res
        else:
            ret = res.get(server, {})
        if mask.has_filters(current_user):
            ret["backups"] = [
                x
                for x in ret.get("backups", [])
                if mask.is_client_allowed(current_user, x.get("name"), server)
            ]
            ret["clients"] = [
                x
                for x in ret.get("clients", [])
                if mask.is_client_allowed(current_user, x.get("name"), server)
            ]
        return ret


@ns.route("/stats", "/<server>/stats", endpoint="clients_stats")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class ClientsStats(Resource):
    """The :class:`burpui.api.clients.ClientsStats` resource allows you to
    access general statistics about your clients.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    client_fields = ns.model(
        "ClientsStatsSingle",
        {
            "last": fields.DateTime(
                required=True, dt_format="iso8601", description="Date of last backup"
            ),
            "last_attempt": fields.DateTime(
                dt_format="iso8601", description="Date of last backup attempt"
            ),
            "name": fields.String(required=True, description="Client name"),
            "state": fields.LocalizedString(
                required=True,
                description="Current state of the client (idle, backup, etc.)",
            ),
            "phase": fields.String(description="Phase of the current running backup"),
            "percent": fields.Integer(description="Percentage done", default=0),
            "labels": fields.List(fields.String, description="List of labels"),
        },
    )

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_list_with(client_fields, code=200, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            500: "Internal failure",
        },
    )
    @browser_cache(1800)
    def get(self, server=None):
        """Returns a list of clients with their states

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              [
                {
                  "last": "2015-05-17 11:40:02",
                  "last_attempt": "2015-05-17 11:40:02",
                  "name": "client1",
                  "state": "idle",
                  "phase": "phase1",
                  "percent": 12,
                  "labels": [
                    "toto"
                  ]
                },
                {
                  "last": "never",
                  "last_attempt": "never",
                  "name": "client2",
                  "state": "idle",
                  "phase": "phase2",
                  "percent": 42,
                  "labels": [
                    "titi"
                  ]
                }
              ]
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above
        """

        server = server or self.parser.parse_args()["serverName"]
        try:
            if (
                not bui.config["STANDALONE"]
                and not current_user.is_anonymous
                and (
                    not current_user.acl.is_admin()
                    and not current_user.acl.is_server_allowed(server)
                )
            ):
                self.abort(403, "Sorry, you don't have any rights on this server")
            jso = bui.client.get_all_clients(agent=server)
            if mask.has_filters(current_user):
                jso = [
                    x
                    for x in jso
                    if mask.is_client_allowed(current_user, x["name"], server)
                ]
        except BUIserverException as e:
            self.abort(500, str(e))
        ret = []
        for client in jso:
            tmp_client = client
            try:
                labels = ClientLabels._get_labels(client["name"], server)
            except BUIserverException as exp:
                self.abort(500, str(exp))
            tmp_client["labels"] = labels
            ret.append(tmp_client)
            if tmp_client["state"] != "idle":
                g.DONOTCACHE = True

        return ret


@ns.route("/all", "/<server>/all", endpoint="clients_all")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
    },
)
class AllClients(Resource):
    """The :class:`burpui.api.clients.AllClients` resource allows you to
    retrieve a list of all clients with their associated server (if any).

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    parser.add_argument(
        "user", help="For which user do we want the data (only works for admins"
    )
    client_fields = ns.model(
        "AllClients",
        {
            "name": fields.String(required=True, description="Client name"),
            "agent": fields.String(
                required=False, default=None, description="Associated Agent name"
            ),
        },
    )

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_list_with(client_fields, code=200, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
        },
    )
    @browser_cache(1800)
    def get(self, server=None):
        """Returns a list of all clients with their associated Agent if any

        **GET** method provided by the webservice.

        The *JSON* returned is:

        ::

            [
              {
                "name": "client1",
                "agent": "agent1"
              },
              {
                "name": "client2",
                "agent": "agent1"
              },
              {
                "name": "client3",
                "agent": "agent2"
              }
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above
        """
        ret = []
        is_admin = True
        args = self.parser.parse_args()
        server = server or args["serverName"]

        is_admin = current_user.is_anonymous or current_user.acl.is_admin()
        is_moderator = current_user.is_anonymous or current_user.acl.is_moderator()

        user = (
            (args.get("user", current_user.name) or current_user.name)
            if is_admin or is_moderator
            else current_user.name
        )

        # drop privileges when switching user
        if user != current_user.name:
            is_admin = False
            is_moderator = False

        if (
            server
            and not is_admin
            and not is_moderator
            and not current_user.acl.is_server_allowed(server)
        ):
            self.abort(403, "You are not allowed to view this server infos")

        if server:
            try:
                clients = [
                    x["name"]
                    for x in bui.client.get_all_clients(
                        agent=server, last_attempt=False
                    )
                ]
            except BUIserverException:
                clients = []
            if not is_admin:
                # use the bui.acl module since we impersonalized the user
                ret = [
                    {"name": x, "agent": server}
                    for x in clients
                    if bui.acl.is_client_allowed(user, x, server)
                ]
            else:
                ret = [{"name": x, "agent": server} for x in clients]
            return ret

        if bui.config["STANDALONE"]:
            try:
                clients = [
                    x["name"] for x in bui.client.get_all_clients(last_attempt=False)
                ]
            except BUIserverException:
                clients = []
            if not is_admin:
                ret = [
                    {"name": x} for x in clients if bui.acl.is_client_allowed(user, x)
                ]
            else:
                ret = [{"name": x} for x in clients]
        else:
            grants = {}
            clients_cache = {}
            for serv in bui.client.servers:
                try:
                    clients = [
                        x["name"]
                        for x in bui.client.get_all_clients(serv, last_attempt=False)
                    ]
                    clients_cache[serv] = clients
                except BUIserverException:
                    clients = []
            if not is_admin:
                for serv in bui.client.servers:
                    grants[serv] = [
                        x
                        for x in clients_cache.get(serv, [])
                        if bui.acl.is_client_allowed(user, x, serv)
                    ]
            else:
                for serv in bui.client.servers:
                    grants[serv] = "all"
            for (serv, clients) in grants.items():
                if not isinstance(clients, list):
                    clients = clients_cache.get(serv, [])
                ret += [{"name": x, "agent": serv} for x in clients]

        return ret



# -*- coding: utf8 -*-
import os
import inspect

from .interface import BUIaudit, BUIauditLogger as BUIauditLoggerInterface

from importlib import import_module
from collections import OrderedDict
from flask_login import current_user


class BUIauditLoader(BUIaudit):
    """See :class:`burpui.misc.audit.interface.BUIaudit`"""

    def __init__(self, app=None):
        """See :func:`burpui.misc.audit.interface.BUIaudit.__init__`

        :param app: Instance of the app we are running in
        :type app: :class:`burpui.engines.server.BUIServer`
        """
        self.app = app
        backends = []
        self.errors = {}
        if self.app.audit_backends and "none" not in self.app.audit_backends:
            me, _ = os.path.splitext(os.path.basename(__file__))
            back = self.app.audit_backends
            for au in back:
                if au == me:
                    self.app.logger.critical("Recursive import not permitted!")
                    continue
                try:
                    (modpath, _) = __name__.rsplit(".", 1)
                    mod = import_module("." + au, modpath)
                    obj = mod.BUIauditLoader(self.app)
                    backends.append(obj)
                except:
                    import traceback

                    self.errors[au] = traceback.format_exc()
        for name, plugin in self.app.plugin_manager.get_plugins_by_type(
            "audit"
        ).items():
            try:
                obj = plugin.BUIauditLoader(self.app)
                backends.append(obj)
            except:
                import traceback

                self.errors[name] = traceback.format_exc()
        backends.sort(key=lambda x: getattr(x, "priority", -1), reverse=True)
        if (
            not backends
            and self.app.audit_backends
            and "none" not in self.app.audit_backends
        ):
            raise ImportError(
                "No backend found for '{}':\n{}".format(
                    self.app.audit_backends, self.errors
                )
            )
        for name, err in self.errors.items():
            self.app.logger.error(
                "Unable to load module {}:\n{}".format(repr(name), err)
            )
        self.backends = OrderedDict()
        for obj in backends:
            self.backends[obj.name] = obj
        self._logger = BUIauditLogger(self)


class BUIauditLogger(BUIauditLoggerInterface):
    def __init__(self, loader):
        self.loader = loader

    def log(self, level, message, *args, **kwargs):
        server_log = ""
        if "server" in kwargs:
            server = kwargs["server"]
            del kwargs["server"]
            if server:
                server_log = f" on {server}"
        if current_user and not current_user.is_anonymous:
            msg = f"{current_user} -> {message}{server_log}"
        else:
            msg = f"{message}{server_log}"
        caller = ""
        stack = inspect.stack()
        exclude = [
            "audit/interface.py",
            "audit/handler.py",
        ]
        for frame in stack:
            if any(frame.filename.endswith(x) for x in exclude):
                continue
            caller = f"{frame.function} [{frame.filename}:{frame.lineno}]"
            break
        if "extra" in kwargs and isinstance(kwargs["extra"], dict):
            kwargs["extra"]["from"] = caller
        else:
            kwargs["extra"] = {"from": caller}
        for back in self.loader.backends.values():
            back.logger.log(level, msg, *args, **kwargs)



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.backup
    :platform: Unix
    :synopsis: Burp-UI backup api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from . import api
from ..engines.server import BUIServer  # noqa
from .custom import Resource
from ..exceptions import BUIserverException

from flask import current_app
from flask_login import current_user

bui = current_app  # type: BUIServer
ns = api.namespace("backup", "Backup methods")


@ns.route(
    "/server-backup/<name>",
    "/<server>/server-backup/<name>",
    methods=["GET", "DELETE"],
    endpoint="is_server_backup",
)
@ns.route(
    "/do-server-backup/<name>",
    "/<server>/do-server-backup/<name>",
    methods=["PUT"],
    endpoint="server_backup",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent mode",
        "name": "Client name",
    },
)
class ServerBackup(Resource):
    """The :class:`burpui.api.backup.ServerBackup` resource allows you to
    prepare a server-initiated backup.

    This resource is part of the :mod:`burpui.api.backup` module.
    """

    @ns.doc(
        responses={
            200: "Success",
            400: "Missing parameter",
            403: "Insufficient permissions",
            500: "Internal failure",
        },
    )
    def get(self, server=None, name=None):
        """Tells if a 'backup' file is present

        **GET** method provided by the webservice.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :returns: True if the file is found
        """
        if not name:
            self.abort(400, "Missing options")
        # Manage ACL
        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_client_allowed(name, server)
        ):
            self.abort(403, "You are not allowed to access this client")
        try:
            return {"is_server_backup": bui.client.is_server_backup(name, server)}
        except BUIserverException as e:
            self.abort(500, str(e))

    @ns.doc(
        responses={
            200: "Success",
            400: "Missing parameter",
            403: "Insufficient permissions",
            500: "Internal failure",
        },
    )
    def delete(self, server=None, name=None):
        """Remove the 'backup' file if present

        **DELETE** method provided by the webservice.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :returns: Status message (success or failure)
        """
        if not name:
            self.abort(400, "Missing options")
        # Manage ACL
        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_moderator()
            and not current_user.acl.is_client_rw(name, server)
        ):
            self.abort(403, "You are not allowed to cancel a backup for this client")
        try:
            bui.audit.logger.info(
                f"requested server-initiated backup cancellation of {name}",
                server=server,
            )
            return bui.client.cancel_server_backup(name, server)
        except BUIserverException as e:
            self.abort(500, str(e))

    @ns.doc(
        responses={
            201: "Success",
            400: "Missing parameter",
            403: "Insufficient permissions",
            500: "Internal failure",
        },
    )
    def put(self, server=None, name=None):
        """Schedule a server-initiated backup

        **PUT** method provided by the webservice.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :returns: Status message (success or failure)
        """
        json = []
        # Check params
        if not name:
            self.abort(400, "Missing options")
        # Manage ACL
        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_moderator()
            and not current_user.acl.is_client_rw(name, server)
        ):
            self.abort(403, "You are not allowed to schedule a backup for this client")
        try:
            bui.audit.logger.info(
                f"requested server-initiated backup scheduling for {name}",
                server=server,
            )
            json = bui.client.server_backup(name, server)
            return json, 201
        except BUIserverException as e:
            self.abort(500, str(e))



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.prefs
    :platform: Unix
    :synopsis: Burp-UI prefs api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from flask import session, current_app, request
from flask_login import current_user
from werkzeug.datastructures import MultiDict

from . import api
from ..engines.server import BUIServer  # noqa
from ..ext.i18n import LANGUAGES
from .custom import fields, Resource

bui = current_app  # type: BUIServer
ns = api.namespace("preferences", "Preferences methods")


@ns.route("/ui/hide", endpoint="prefs_ui_hide")
class PrefsUIHide(Resource):
    """The :class:`burpui.api.prefs.PrefsUI` resource allows you to
    set your UI preferences.

    This resource is part of the :mod:`burpui.api.prefs` module.
    """

    parser = ns.parser()
    parser.add_argument("name", dest="client", help="Client to hide")
    parser.add_argument("agent", dest="server", help="Server to hide")

    hidden_model = ns.model(
        "HiddenModel",
        {
            "client": fields.String(description="Hidden client name"),
            "server": fields.String(description="Hidden server name"),
        },
    )

    @ns.marshal_list_with(hidden_model, description="Success", code=200)
    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
        },
    )
    def get(self):
        """Returns a list of hidden clients/servers

        **GET** method provided by the webservice.

        :returns: list
        """
        if (
            bui.config["WITH_SQL"]
            and not bui.config["BUI_DEMO"]
            and not current_user.is_anonymous
        ):
            from ..models import Hidden

            return Hidden.query.filter_by(user=current_user.name).all()
        return []

    @ns.expect(parser)
    @ns.marshal_with(hidden_model, description="Success", code=200)
    @ns.doc(
        responses={
            200: "Success, object not recorder",
            201: "Success, object recorded",
            403: "Insufficient permissions",
            500: "Internal server error",
        },
    )
    def put(self):
        """Hide a client/server

        **PUT** method provided by the webservice.

        :returns: Object to hide
        """
        ret = []
        args = self.parser.parse_args()
        if (
            bui.config["WITH_SQL"]
            and not bui.config["BUI_DEMO"]
            and not current_user.is_anonymous
        ):
            from ..ext.sql import db
            from ..models import Hidden

            client = args.get("client") or None
            server = args.get("server") or None
            hidden = Hidden.query.filter_by(
                client=client, server=server, user=current_user.name
            ).first()
            if not hidden:
                hide = Hidden(current_user.name, client, server)
                db.session.add(hide)
                try:
                    db.session.commit()
                except:  # pragma: no cover
                    db.session.rollback()
                    self.abort(500, "Internal server error")
                return hide, 201
            return hidden
        return ret

    @ns.expect(parser)
    @ns.doc(
        responses={
            204: "Success",
            403: "Insufficient permissions",
            500: "Internal server error",
        },
    )
    def delete(self):
        """Make a client/server visible again

        **DELETE** method provided by the webservice.
        """
        args = self.parser.parse_args()
        if (
            bui.config["WITH_SQL"]
            and not bui.config["BUI_DEMO"]
            and not current_user.is_anonymous
        ):
            from ..ext.sql import db
            from ..models import Hidden

            hide = Hidden.query.filter_by(
                client=(args.get("client") or None),
                server=(args.get("server") or None),
                user=current_user.name,
            ).first()
            if hide:
                db.session.delete(hide)
                try:
                    db.session.commit()
                except:  # pragma: no cover
                    db.session.rollback()
                    self.abort(500, "Internal server error")
        return None, 204


@ns.route("/ui", endpoint="prefs_ui")
class PrefsUI(Resource):
    """The :class:`burpui.api.prefs.PrefsUI` resource allows you to
    set your UI preferences.

    This resource is part of the :mod:`burpui.api.prefs` module.
    """

    parser = ns.parser()
    parser.add_argument(
        "pageLength", type=int, required=False, help="Number of element per page"
    )
    parser.add_argument(
        "language",
        type=str,
        required=False,
        help="Language",
        choices=list(LANGUAGES.keys()),
    )
    parser.add_argument("dateFormat", type=str, required=False, help="Date format")
    parser.add_argument("timezone", type=str, required=False, help="Timezone")

    @staticmethod
    def _user_language(language):
        """Set the current user language"""
        if current_user and not current_user.is_anonymous and language:
            setattr(current_user, "language", language)

    def _store_prefs(self, key, val):
        """Store the prefs if persistent storage is enabled"""
        if bui.config["WITH_SQL"] and not bui.config["BUI_DEMO"]:
            from ..ext.sql import db
            from ..models import Pref

            pref = Pref.query.filter_by(user=current_user.name, key=key).first()
            if pref:
                if val:
                    pref.value = val
                else:
                    db.session.delete(pref)
            elif val:
                pref = Pref(current_user.name, key, val)
                db.session.add(pref)
            try:
                db.session.commit()
            except:  # pragma: no cover
                db.session.rollback()

    def _update_prefs(self):
        """update prefs"""
        args = self.parser.parse_args()
        sess = session._get_current_object()
        ret = {}
        req = MultiDict()
        data = getattr(request, "values", None)
        if data:
            req.update(data)
        data = request.get_json(silent=True)
        if data:
            req.update(data)
        for key in args.keys():
            if key not in req:
                continue
            temp = args.get(key)
            if temp:
                if key == "language":
                    self._user_language(temp)
                sess[key] = temp
            elif key in sess:  # pragma: no cover
                del sess[key]
            ret[key] = temp
            self._store_prefs(key, temp)

        return ret

    @ns.doc(
        responses={
            200: "Success",
            403: "Insufficient permissions",
        },
    )
    def get(self):
        """Returns a list of prefs

        **GET** method provided by the webservice.

        :returns: prefs
        """
        args = self.parser.parse_args()
        ret = {}
        sess = session
        for key in args.keys():
            ret[key] = sess.get(key)
        return ret

    @ns.expect(parser, validate=True)
    @ns.doc(
        responses={
            201: "Success",
            403: "Not allowed",
            400: "Missing parameters",
        },
    )
    def put(self):
        """Create prefs"""
        return self._update_prefs(), 201

    @ns.expect(parser)
    @ns.doc(
        responses={
            200: "Success",
            403: "Not allowed",
            400: "Missing parameters",
        },
    )
    def delete(self):
        """Delete prefs"""
        args = self.parser.parse_args()
        sess = session
        ret = {}
        for key in args.keys():
            temp = args.get(key)
            if temp:
                del sess[key]
                if bui.config["WITH_SQL"]:
                    from ..ext.sql import db
                    from ..models import Pref

                    try:
                        Pref.query.filter_by(user=current_user.name, key=key).delete()
                        db.session.commit()
                    except:  # pragma: no cover
                        db.session.rollback()
            ret[key] = sess.get(key)

        return ret

    @ns.expect(parser, validate=True)
    @ns.doc(
        responses={
            200: "Success",
            403: "Not allowed",
            400: "Missing parameters",
        },
    )
    def post(self):
        """Change prefs"""
        return self._update_prefs()



# -*- coding: utf8 -*-
"""
.. module:: burpui.config
    :platform: Unix
    :synopsis: Burp-UI config module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import re
import codecs
import configobj
import validate

from .tools.logging import logger


class BUIConfig(dict):
    """Custom config parser"""

    logger = logger
    mtime = 0

    def __init__(self, config=None, defaults=None):
        """Wrapper around the ConfigObj class

        :param config: Configuration to parse
        :type config: str, list or File

        :param defaults: Default options
        :type defaults: dict
        """
        if defaults is not None:
            self.defaults = defaults
        if config:
            self.parse(config, defaults)

    def parse(self, config, defaults=None):
        """Parse the conf

        :param config: Configuration to parse
        :type config: str, list or File

        :param defaults: Default options
        :type defaults: dict
        """
        self.conf = {}
        self.conffile = config
        self.section = None
        if defaults is not None or not hasattr(self, "defaults"):
            self.defaults = defaults
        self.validator = validate.Validator()
        try:
            self.conf = configobj.ConfigObj(config, encoding="utf-8")
            self.mtime = os.path.getmtime(self.conffile)
        except configobj.ConfigObjError as exp:
            # We were unable to parse the config
            self.logger.critical("Unable to parse configuration")
            raise exp

    @property
    def options(self):
        """ConfigObj object"""
        self._refresh()
        return self.conf

    @property
    def id(self):
        """Conf ID to detect changes"""
        return self.mtime

    @staticmethod
    def string_lower_list(value):
        if not value:
            raise validate.VdtMissingValue("No value for this option")
        if not isinstance(value, list):
            return [str(value).lower()]
        return [str(x).lower() for x in value]

    @staticmethod
    def force_string(value):
        if not value:
            raise validate.VdtMissingValue("No value for this option")
        if not isinstance(value, list):
            return str(value)
        return ",".join(value)

    def boolean_or_string(self, value):
        if not value:
            raise validate.VdtMissingValue("No value for this option")
        try:
            caster = self.validator.functions.get("boolean")
            return caster(value)
        except validate.VdtTypeError:
            return value

    def lookup_section(self, section, source=None):
        """Lookup for a given section. If the section is not found in the conf,
        we try to search it in the comments to uncomment it.
        If it is missing from the conf and the comments, we append it in the
        conf.
        """
        ret = True
        if not self.section_exists(section):
            # look for the section in the comments
            conffile = self.options.filename
            source = source or conffile
            ori = []
            with codecs.open(source, "r", "utf-8", errors="ignore") as config:
                ori = [x.rstrip("\n") for x in config.readlines()]
            if ori:
                with codecs.open(conffile, "w", "utf-8", errors="ignore") as config:
                    found = False
                    for line in ori:
                        if re.match(r"^\s*(#|;)+\s*\[{}\]".format(section), line):

                            config.write("[{}]\n".format(section))
                            found = True
                        else:
                            config.write("{}\n".format(line))

                    if not found:
                        config.write("[{}]\n".format(section))
                ret = False
        return ret

    def section_exists(self, section):
        """Check whether a section exists or not"""
        return section in self.options

    def rename_section(self, old_section, new_section, source=None):
        """Rename a given section"""
        ret = False
        if not self.section_exists(old_section):
            return ret
        conffile = self.options.filename
        source = source or conffile
        ori = []
        with codecs.open(source, "r", "utf-8", errors="ignore") as config:
            ori = [x.rstrip("\n") for x in config.readlines()]
        if ori:
            with codecs.open(conffile, "w", "utf-8", errors="ignore") as config:
                for line in ori:
                    if re.match(r"^\s*(#|;)*\s*\[{}\]".format(old_section), line):
                        config.write(
                            "{}\n".format(line.replace(old_section, new_section))
                        )
                        ret = True
                    else:
                        config.write("{}\n".format(line))
        return ret

    def _rename_option_full(self, orig_option, dest_option, orig_section, dest_section):
        """Rename a given option and possibly moves it to another section
        :return: True if the option have been successfully renamed/moved
        :rtype: bool

        :raises ValueError: if the ``orig_section`` does not exist
        :raises KeyError: if the ``orig_option`` does not exist in the ``orig_section``
        """
        if not self.section_exists(orig_section):
            raise ValueError("No such section: {}".format(orig_section))

        orig = self.conf[orig_section]
        if orig_option not in orig:
            raise KeyError(
                "No such option in the [{}] section: {}".format(
                    orig_section, orig_option
                )
            )

        # adding the new section if it is missing
        if orig_section != dest_section and not self.lookup_section(dest_section):
            self._refresh(True)

        dest = self.conf[dest_section]
        comments = orig.comments[orig_option]
        inline_comments = orig.inline_comments[orig_option]

        # copy value and comments from orig to dest
        dest[dest_option] = orig[orig_option]
        dest.comments[dest_option] = comments
        dest.inline_comments[dest_option] = inline_comments

        # remove orig key
        del orig[orig_option]

        # save
        self.conf.write()
        return True

    def rename_option(self, orig_option, dest_option, section):
        """Rename a given option"""
        # this is useless
        if orig_option == dest_option:
            return False
        return self._rename_option_full(orig_option, dest_option, section, section)

    def move_option(self, option, orig_section, dest_section):
        """Move an option to another section, if you need to rename the option use the
        _rename_option_full function instead"""
        # useless
        if orig_section == dest_section:
            return False
        return self._rename_option_full(option, option, orig_section, dest_section)

    def changed(self, id):
        """Check if the conf has changed"""
        # don't use delta for cases where we run several gunicorn workers
        self._refresh()
        return id != self.mtime

    def _refresh(self, force=False):
        """Refresh conf"""
        mtime = os.path.getmtime(self.conffile)
        if mtime != self.mtime or force:
            self.logger.debug("Configuration changed")
            self.mtime = mtime
            self.conf.reload()

    def update_defaults(self, new_defaults):
        """Add new defaults"""
        self.defaults.update(new_defaults)

    def default_section(self, section):
        """Set the default section"""
        self.section = section

    def safe_get(self, key, cast="pass", section=None, defaults=None):
        """Safely return the asked option

        :param key: Key name
        :type key: str

        :param cast: How to cast the option
        :type cast: str

        :param section: Section name
        :type section: str

        :param defaults: Default options
        :type defaults: dict, mixed

        :returns: The value of the asked option
        """
        # The configobj validator is sooo broken. We need to workaround it...
        default_by_type = {
            "integer": 0,
            "boolean": False,
        }
        section = section or self.section
        # if the defaults argument is not a dict, assume it's a single value
        if defaults and not isinstance(defaults, dict):
            defaults = {section: {key: defaults}}
        defaults = defaults or self.defaults
        if section not in self.conf:
            self.logger.warning("No '{}' section found".format(section))
            if defaults:
                return defaults.get(section, {}).get(key)
            return None

        val = self.conf.get(section).get(key)
        default = default_by_type.get(cast)
        if defaults and section in defaults and key in defaults.get(section):
            default = defaults.get(section, {}).get(key)
        try:
            caster = self.validator.functions.get(cast)
            if not caster:
                try:
                    caster = getattr(self, cast)
                except AttributeError:
                    self.logger.error("'{}': no such validator".format(cast))
                    return val
            if cast == "force_list" and val is None:
                val = []
            ret = caster(val)
            # special case for boolean and integer, etc.
            if ret is None:
                ret = default
            self.logger.debug(
                "[{}]:{} - found: {}, default: {} -> {}".format(
                    section, key, val, default, ret
                )
            )
        except validate.ValidateError as exp:
            ret = default
            self.logger.info(
                "{}\n[{}]:{} - found: {}, default: {} -> {}".format(
                    str(exp), section, key, val, default, ret
                )
            )
        return ret


config = BUIConfig()



# -*- coding: utf8 -*-
"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.app
    :platform: Unix
    :synopsis: Burp-UI app module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import sys
import json
import time
import logging

from .desc import __version__, __release__
from .extensions import (
    create_celery,
    create_db,
    create_websocket,
    parse_db_setting,
    get_redis_server,
)


def create_app(conf=None, verbose=0, logfile=None, **kwargs):
    """Initialize the whole application.

    :param conf: Configuration file to use
    :type conf: str

    :param verbose: Set the verbosity level
    :type verbose: int

    :param logfile: Store the logs in the given file
    :type logfile: str

    :param kwargs: Extra options:
                   - gunicorn (bool): Enable gunicorn engine instead of flask's
                   default. Default is True.
                   - unittest (bool): Are we running tests (used for test only).
                   Default is False.
                   - debug (bool): Enable debug mode. Default is False.
                   - cli (bool): Are we running the CLI. Default is False.
                   - reverse_proxy (bool): Are we behind a reverse-proxy.
                   Default is True if gunicorn is True
                   - websocket_server (bool): Are we running the websocket
                   server. Default is False
    :type kwargs: dict

    :returns: A :class:`burpui.engines.server.BUIServer` object
    """
    from flask import g, request, session
    from flask_login import LoginManager
    from flask_babel import gettext
    from .thirdparty.flask_bower import Bower
    from .utils import ReverseProxied, lookup_file, is_uuid
    from .tools.logging import logger
    from .security import basic_login_from_request
    from .engines.server import BUIServer as BurpUI
    from .sessions import session_manager
    from .filter import mask
    from .ext.cache import cache
    from .ext.i18n import babel, get_locale
    from .misc.auth.handler import BUIanon

    gunicorn = kwargs.get("gunicorn", True)
    unittest = kwargs.get("unittest", False)
    debug = kwargs.get("debug", False)
    cli = kwargs.get("cli", False)
    reverse_proxy = kwargs.get("reverse_proxy", gunicorn)
    celery_worker = kwargs.get("celery_worker", False)
    websocket_server = kwargs.get("websocket_server", False)

    # We initialize the core
    app = BurpUI()

    app.config["CFG"] = None
    app.config["LOG_FILE"] = logfile
    app.config["LOG_LEVEL"] = verbose

    logger.init_app(app)
    if verbose:
        app.enable_logger()

    app.gunicorn = gunicorn

    logger.debug(
        "conf: {}\n".format(conf)
        + "verbose: {}\n".format(logging.getLevelName(verbose))
        + "logfile: {}\n".format(logfile)
        + "gunicorn: {}\n".format(gunicorn)
        + "debug: {}\n".format(debug)
        + "unittest: {}\n".format(unittest)
        + "cli: {}\n".format(cli)
        + "reverse_proxy: {}\n".format(reverse_proxy)
        + "celery_worker: {}\n".format(celery_worker)
        + "websocket_server: {}".format(websocket_server)
    )

    # Some config
    app.config["BUI_CLI"] = cli

    # FIXME: strange behavior when bundling errors
    # app.config['BUNDLE_ERRORS'] = True

    app.config["REMEMBER_COOKIE_HTTPONLY"] = True

    if debug and not gunicorn:  # pragma: no cover
        app.config["DEBUG"] = True and not unittest
        app.config["TESTING"] = True and not unittest

    # Still need to test conf file here because the init function can be called
    # by gunicorn directly
    if conf:
        app.config["CFG"] = lookup_file(conf, guess=False)
    else:
        app.config["CFG"] = lookup_file()

    logger.info("Using configuration: {}".format(app.config["CFG"]))

    app.setup(app.config["CFG"], unittest, cli)
    if cli and not websocket_server and "shell" not in sys.argv:
        return app

    if debug:
        app.config.setdefault("TEMPLATES_AUTO_RELOAD", True)
        app.config["TEMPLATES_AUTO_RELOAD"] = True
        app.config["DEBUG"] = True

    if app.demo:
        try:
            import sentry_sdk
            from sentry_sdk.integrations.flask import FlaskIntegration

            sentry_sdk.init(
                dsn=app.config["BUI_DSN"], integrations=[FlaskIntegration()]
            )
        except ImportError:
            pass

    # manage application secret key
    if app.secret_key and (
        app.secret_key.lower() == "none"
        or (app.secret_key.lower() == "random" and gunicorn)
    ):  # pragma: no cover
        logger.critical(
            "Your setup is not secure! Please consider setting a"
            " secret key in your configuration file"
        )
        app.secret_key = "Burp-UI"
    if not app.secret_key or app.secret_key.lower() == "random":
        from base64 import b64encode

        app.secret_key = b64encode(os.urandom(256))

    app.wsgi_app = ReverseProxied(app.wsgi_app, app)

    # Manage reverse_proxy special tricks & improvements
    if reverse_proxy:  # pragma: no cover
        from werkzeug.middleware.proxy_fix import ProxyFix

        kwargs = {}
        if app.config["NUM_PROXIES"] > 0:
            kwargs = app.config["PROXY_FIX_ARGS"].format(
                num_proxies=app.config["NUM_PROXIES"]
            )
            kwargs = json.loads(kwargs)
        logger.debug(f"Using {kwargs} as ProxyFix parameters")
        app.wsgi_app = ProxyFix(app.wsgi_app, **kwargs)

    if app.storage and app.storage.lower() == "redis":
        try:
            # Session setup
            if not app.session_db or str(app.session_db).lower() not in [
                "none",
                "false",
            ]:
                from redis import Redis
                from .ext.session import sess

                host, port, pwd = get_redis_server(app)
                db = 0
                if app.session_db and str(app.session_db).lower() not in [
                    "redis",
                    "default",
                    "true",
                ]:
                    try:  # pragma: no cover
                        (_, _, pwd, host, port, db) = parse_db_setting(app.session_db)
                    except ValueError as exp:
                        logger.warning(str(exp))
                try:
                    db = int(db)
                except ValueError:
                    db = 0
                logger.debug(
                    "SESSION: Using redis://guest:****@{}:{}/{}".format(host, port, db)
                )
                red = Redis(host=host, port=port, db=db, password=pwd)
                app.config["WITH_SRV_SESSION"] = True
                app.config["SESSION_TYPE"] = "redis"
                app.config["SESSION_REDIS"] = red
                app.config["SESSION_USE_SIGNER"] = app.secret_key is not None
                app.config["SESSION_PERMANENT"] = False
                sess.init_app(app)
                session_manager.backend = red
        except Exception as exp:  # pragma: no cover
            logger.warning("Unable to initialize session: {}".format(str(exp)))
            app.config["WITH_SRV_SESSION"] = False
        try:
            # Cache setup
            if not app.cache_db or str(app.cache_db).lower() not in ["none", "false"]:
                host, port, pwd = get_redis_server(app)
                db = 1
                if app.cache_db and str(app.cache_db).lower() not in [
                    "redis",
                    "default",
                    "true",
                ]:
                    try:  # pragma: no cover
                        (_, _, pwd, host, port, db) = parse_db_setting(app.cache_db)
                    except ValueError as exp:
                        logger.warning(str(exp))
                try:
                    db = int(db)
                except ValueError:
                    db = 1
                logger.debug(
                    "CACHE: Using redis://guest:****@{}:{}/{}".format(host, port, db)
                )
                cache.init_app(
                    app,
                    config={
                        "CACHE_TYPE": "redis",
                        "CACHE_REDIS_HOST": host,
                        "CACHE_REDIS_PORT": port,
                        "CACHE_REDIS_PASSWORD": pwd,
                        "CACHE_REDIS_DB": db,
                    },
                )
                # clear cache at startup in case we removed or added servers
                with app.app_context():
                    cache.clear()
            else:  # pragma: no cover
                cache.init_app(app)
        except Exception as exp:  # pragma: no cover
            logger.warning("Unable to initialize cache: {}".format(str(exp)))
            cache.init_app(app)
        try:
            # Limiter setup
            if app.limiter and str(app.limiter).lower() not in [
                "none",
                "false",
            ]:  # pragma: no cover
                from .ext.limit import limiter

                app.config["RATELIMIT_HEADERS_ENABLED"] = True
                if app.limiter and str(app.limiter).lower() not in [
                    "default",
                    "redis",
                    "true",
                ]:
                    app.config["RATELIMIT_STORAGE_URL"] = app.limiter
                else:
                    db = 3
                    host, port, pwd = get_redis_server(app)
                    if pwd:
                        conn = "redis://guest:{}@{}:{}/{}".format(pwd, host, port, db)
                    else:
                        conn = "redis://{}:{}/{}".format(host, port, db)
                    app.config["RATELIMIT_STORAGE_URL"] = conn

                (_, _, pwd, host, port, db) = parse_db_setting(
                    app.config["RATELIMIT_STORAGE_URL"]
                )

                logger.debug(
                    "LIMITER: Using redis://guest:****@{}:{}/{}".format(host, port, db)
                )
                limiter.init_app(app)
                app.config["WITH_LIMIT"] = True
        except ImportError:  # pragma: no cover
            logger.warning(
                "Unable to load limiter. Did you run 'pip install " "flask-limiter'?"
            )
        except Exception as exp:  # pragma: no cover
            logger.warning("Unable to initialize limiter: {}".format(str(exp)))
    else:
        cache.init_app(app)

    # Initialize i18n
    babel.init_app(app)

    # Create SQLAlchemy if enabled
    create_db(app, cli, unittest, celery_worker=celery_worker)

    if not celery_worker:
        from .api import api, apibp
        from .routes import view, mypad

        app.jinja_env.globals.update(
            isinstance=isinstance,
            list=list,
            mypad=mypad,
            version_id="{}-{}".format(__version__, __release__),
        )

        # We initialize the API
        api.load_all()
        app.register_blueprint(apibp)

        # Then we load our routes
        app.register_blueprint(view)

        # Initialize Bower ext
        app.config.setdefault("BOWER_COMPONENTS_ROOT", os.path.join("static", "vendor"))
        app.config.setdefault("BOWER_REPLACE_URL_FOR", True)
        bower = Bower()
        bower.init_app(app)

    # Order of the initialization matters!
    # The websocket must be configured prior to the celery worker for instance

    # Initialize Session Manager
    session_manager.init_app(app)
    # Initialize filter
    mask.init_app(app)

    # And the login_manager
    app.login_manager = LoginManager()
    app.login_manager.anonymous_user = BUIanon
    app.login_manager.login_view = "view.login"
    app.login_manager.login_message_category = "info"
    app.login_manager.session_protection = "strong"
    # This is just to have the strings in the .po files
    app.login_manager.login_message = gettext("Please log in to access this page.")
    app.login_manager.needs_refresh_message = gettext(
        "Please reauthenticate to access this page."
    )
    # This will be called at runtime and will then translate the strings
    app.login_manager.localize_callback = gettext
    app.login_manager.init_app(app)

    # Create WebSocket server
    if create_websocket(app, websocket_server, celery_worker, cli):
        return app

    # Create celery app if enabled
    create_celery(app, warn=False)

    def _check_session(user, request, api=False):
        """Check if the session is in the db"""
        if user and not session_manager.session_in_db():  # pragma: no cover
            login = getattr(user, "name", None)
            if login and not is_uuid(login):
                remember = session.get("persistent", False)
                if not remember:
                    from flask_login import decode_cookie

                    remember_cookie = request.cookies.get(
                        app.config.get("REMEMBER_COOKIE_NAME"), False
                    )
                    # check if the remember_cookie is legit
                    if remember_cookie and decode_cookie(remember_cookie):
                        remember = True
                session_manager.store_session(
                    login,
                    request.remote_addr,
                    request.headers.get("User-Agent"),
                    remember,
                    api,
                )
            elif login:
                app.uhandler.remove(login)

    @app.before_request
    def setup_request():
        g.version = "{}-{}".format(__version__, __release__)
        g.locale = get_locale()
        g.now = round(time.time())
        g.date_format = session.get("dateFormat", "llll")
        g.timezone = session.get("timezone")
        # make sure to store secure cookie if required
        if app.config["BUI_SCOOKIE"]:
            criteria = (
                request.is_secure,
                request.headers.get("X-Forwarded-Proto", "http") == "https",
            )
            app.config["SESSION_COOKIE_SECURE"] = app.config[
                "REMEMBER_COOKIE_SECURE"
            ] = any(criteria)
        if "_extra" in request.args:
            session["_extra"] = request.args.get("_extra")
        g._extra = session.get("_extra", "")

    @app.login_manager.user_loader
    def load_user(userid):
        """User loader callback"""
        if app.auth != "none":
            user = app.uhandler.user(userid)
            if not user:
                return None
            if "X-Language" in request.headers:
                language = request.headers.get("X-Language")
                user.language = language
                session["language"] = language
            if "_id" not in session:
                from flask_login import login_user

                # if _id is not in session, it means we loaded the user from
                # cache/db using the remember cookie so we need to login it
                login_user(user, remember=user.is_authenticated, fresh=False)
            _check_session(user, request)
            return user
        return None

    @app.login_manager.request_loader
    def load_user_from_request(request):
        """User loader from request callback"""
        if app.auth != "none":
            user = basic_login_from_request(request, app)
            _check_session(user, request, True)
            return user

    @app.after_request
    def after_request(response):
        if getattr(g, "basic_session", False):
            if session_manager.invalidate_current_session():
                session_manager.delete_session()
        return response

    return app



# -*- coding: utf8 -*-
"""
.. module:: burpui.plugins
    :platform: Unix
    :synopsis: Burp-UI plugins module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from pluginbase import PluginBase


class PluginManager(object):
    """The :class:`burpui.plugins.PluginManager` class is a plugin manager.

    :param app: Instance of the app we are running in
    :type app: :class:`burpui.engines.server.BUIServer`

    :param searchpath: The places to look for plugins
    :type searchpath: list
    """

    def __init__(self, app, searchpath):
        self.app = app
        self.searchpath = searchpath
        self.init = False
        self.loaded = False
        self.plugins = {}

    def _init_manager(self):
        if self.init:
            return
        self.plugin_base = PluginBase(package="burpui.plugins.ext")
        self.plugin_source = self.plugin_base.make_plugin_source(
            searchpath=self.searchpath
        )
        self.init = True

    def load_all(self, force=False):
        if self.loaded and not force:
            return
        self._init_manager()
        for plugin_name in self.plugin_source.list_plugins():
            if plugin_name not in self.plugins:
                try:
                    plugin = self.plugin_source.load_plugin(plugin_name)
                    current_type = getattr(plugin, "__type__", None)
                    if not current_type:
                        self.app.logger.warning(
                            "No __type__ for {}. Ignoring it".format(repr(plugin_name))
                        )
                        continue
                    self.app.logger.info(
                        "Loading plugin {} ({})".format(repr(plugin_name), current_type)
                    )
                    self.plugins[plugin_name] = plugin
                except Exception as exp:
                    self.app.logger.error(
                        "Unable to load plugin {}: {}".format(
                            repr(plugin_name), str(exp)
                        )
                    )
        self.loaded = True

    def get_plugins_by_type(self, plugin_type):
        ret = {}
        for name, plugin in self.plugins.items():
            current_type = getattr(plugin, "__type__", None)
            if not current_type:
                self.app.logger.warning(
                    "No __type__ for {}. Ignoring it".format(repr(name))
                )
                continue
            if current_type == plugin_type:
                ret[name] = plugin

        return ret

    def get_plugin_by_name(self, name):
        return self.plugins.get(name, None)



#!/usr/bin/env python
# -*- coding: utf8 -*-
"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.worker
    :platform: Unix
    :synopsis: Burp-UI celery worker module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import sys

# Try to load modules from our current env first
sys.path.insert(
    0, os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..")
)


def init_app(conf):
    from burpui import create_app
    from burpui.ext.tasks import celery

    app = create_app(conf, celery_worker=True, gunicorn=False)
    return app, celery


config = os.getenv("BUI_CONFIG")
app, celery = init_app(config)
app.app_context().push()

if not app.config["WITH_CELERY"]:
    message = (
        "Something went wrong while initializing celery worker.\n"
        "Maybe it is not enabled in your conf ({}).".format(config)
    )
    raise Exception(message)



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.parser.burp2
    :platform: Unix
    :synopsis: Burp-UI configuration file parser for Burp2.
.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
from .burp1 import Parser as Burp1
from ..backend.utils.constant import BURP_LISTEN_OPTION, BURP_BIND_MULTIPLE


def __(string):
    """dummy function to fake the translation"""
    return string


# inherit Burp1 parser so we can just override available options
class Parser(Burp1):
    """Extends :class:`burpui.misc.parser.burp1.Parser`"""

    pver = 2

    _pair_srv = None
    _pair_associations = None
    _multi_srv = None
    _integer_srv = None

    @property
    def pair_associations(self):
        if self._pair_associations is None:
            if (
                self.backend
                and (
                    getattr(self.backend, "server_version", self._assume_server_version)
                    or getattr(
                        self.backend, "client_version", self._assume_client_version
                    )
                    or ""
                )
                >= BURP_LISTEN_OPTION
            ):
                self._pair_associations = {
                    "listen": "max_children",
                    "max_children": "listen",
                    "listen_status": "max_status_children",
                    "max_status_children": "listen_status",
                }
            elif (
                self.backend
                and (
                    getattr(self.backend, "server_version", self._assume_server_version)
                    or getattr(
                        self.backend, "client_version", self._assume_client_version
                    )
                    or ""
                )
                >= BURP_BIND_MULTIPLE
            ):
                self._pair_associations = {
                    "port": "max_children",
                    "max_children": "port",
                    "status_port": "max_status_children",
                    "max_status_children": "status_port",
                }
            else:
                self._pair_associations = {}
        return self._pair_associations

    @property
    def pair_srv(self):
        if self._pair_srv is None:
            if (
                self.backend
                and (
                    getattr(self.backend, "server_version", self._assume_server_version)
                    or getattr(
                        self.backend, "client_version", self._assume_client_version
                    )
                    or ""
                )
                >= BURP_LISTEN_OPTION
            ):
                self._pair_srv = [
                    "listen",
                    "max_children",
                    "listen_status",
                    "max_status_children",
                ]
            elif (
                self.backend
                and (
                    getattr(self.backend, "server_version", self._assume_server_version)
                    or getattr(
                        self.backend, "client_version", self._assume_client_version
                    )
                    or ""
                )
                >= BURP_BIND_MULTIPLE
            ):
                self._pair_srv = [
                    "port",
                    "max_children",
                    "status_port",
                    "max_status_children",
                ]
            else:
                self._pair_srv = []
        return self._pair_srv

    @property
    def multi_srv(self):
        if self._multi_srv is None:
            self._multi_srv = Burp1.multi_srv + ["label"]
            if (
                self.backend
                and (
                    getattr(self.backend, "server_version", self._assume_server_version)
                    or getattr(
                        self.backend, "client_version", self._assume_client_version
                    )
                    or ""
                )
                >= BURP_BIND_MULTIPLE
            ):
                self._multi_srv += ["port", "status_port"]
        return self._multi_srv

    @property
    def integer_srv(self):
        if self._integer_srv is None:
            self._integer_srv = Burp1.integer_srv
            if (
                self.backend
                and (
                    getattr(self.backend, "server_version", self._assume_server_version)
                    or getattr(
                        self.backend, "client_version", self._assume_client_version
                    )
                    or ""
                )
                >= BURP_BIND_MULTIPLE
            ):
                for rem in [
                    "port",
                    "max_children",
                    "status_port",
                    "max_status_children",
                ]:
                    self._integer_srv.remove(rem)
        return self._integer_srv

    advanced_type = Burp1.advanced_type
    advanced_type.update(
        {
            "port": "integer",
            "max_children": "integer",
            "status_port": "integer",
            "max_status_children": "integer",
        }
    )
    string_srv = Burp1.string_srv + [
        "manual_delete",
        "rblk_memory_max",
    ]
    boolean_add = [
        "acl",
        "xattr",
        "glob_after_script_pre",
        "cname_fqdn",
        "cname_lowercase",
    ]
    boolean_add_cli = [
        "enabled",
    ]
    boolean_srv = Burp1.boolean_srv + boolean_add
    boolean_cli = Burp1.boolean_cli + boolean_add + boolean_add_cli
    multi_cli = Burp1.multi_cli + [
        "label",
    ]
    integer_cli = Burp1.integer_cli + [
        "randomise",
    ]
    fields_cli = (
        Burp1.fields_cli
        + boolean_add
        + boolean_add_cli
        + [
            "randomise",
            "manual_delete",
            "label",
        ]
    )
    placeholders = Burp1.placeholders
    placeholders.update(
        {
            "acl": "0|1",
            "xattr": "0|1",
            "randomise": __("max secs"),
            "manual_delete": __("path"),
            "label": __("some informations"),
            "listen": __("[address]:[port]"),
            "listen_status": __("[address]:[port]"),
            "status_address": __("address|localhost"),
            "glob_after_script_pre": "0|1",
            "enabled": "0|1",
            "cname_fqdn": "0|1",
            "cname_lowercase": "0|1",
            "rblk_memory_max": "b/Kb/Mb/Gb",
        }
    )
    values = Burp1.values
    # status_address can now listen on any address
    del values["status_address"]
    defaults = Burp1.defaults
    defaults.update(
        {
            "acl": True,
            "xattr": True,
            "glob_after_script_pre": True,
            "randomise": 0,
            "manual_delete": "",
            "label": "",
            "enabled": True,
            "cname_fqdn": True,
            "cname_lowercase": False,
            "rblk_memory_max": "256Mb",
        }
    )
    doc = Burp1.doc
    doc.update(
        {
            "acl": __(
                "If acl support is compiled into burp, this allows you to"
                " decide whether or not to backup acls at runtime. The"
                " default is '1'."
            ),
            "xattr": __(
                "If xattr support is compiled into burp, this allows you"
                " to decide whether or not to backup xattrs at runtime."
                " The default is '1'."
            ),
            "randomise": __(
                "When running a timed backup, sleep for a random"
                " number of seconds (between 0 and the number given)"
                " before contacting the server. Alternatively, this"
                " can be specified by the '-q' command line option."
            ),
            "manual_delete": __(
                "This can be overridden by the clientconfdir"
                " configuration files in clientconfdir on the"
                " server. When the server needs to delete old"
                " backups, or rubble left over from generating"
                " reverse patches with librsync=1, it will"
                " normally delete them in place. If you use the"
                " 'manual_delete' option, the files will be moved"
                " to the path specified for deletion at a later"
                " point. You will then need to configure a cron"
                " job, or similar, to delete the files yourself."
                " Do not specify a path that is not on the same"
                " filesystem as the client storage directory."
            ),
            "label": __(
                "You can have multiple labels, and they can be"
                " overridden in the client configuration files in"
                " clientconfdir on the server. They will appear as an"
                " array of strings in the server status monitor JSON"
                " output. The idea is to provide a mechanism for"
                " arbitrary values to be passed to clients of the server"
                " status monitor."
            ),
            "listen": __(
                "Defines the main TCP address and port that the server listens"
                " on. The default is either '::' or '0.0.0.0', dependent upon"
                " compile time options. Specify multiple 'listen' entries on"
                " separate lines in order to listen on multiple addresses and"
                " ports. Each pair can be configured with its own 'max_children'"
                " value."
            ),
            "listen_status": __(
                "Defines the main TCP address and port that the server"
                " listens on for status requests. Specify multiple"
                " 'listen_status' entries on separate lines in order to"
                " listen on multiple addresses and ports. Each pair can"
                " be configured with its own 'max_status_children' value."
                " Comment out to have no status server."
            ),
            "status_address": __(
                "Defines the main TCP address that the server "
                "listens on for status requests. The default  "
                "is  special  value  'localhost'  that includes "
                "both '::1' (if available) and '127.0.0.1' "
                "(always)."
            ),
            "glob_after_script_pre": __(
                "Set this to 0 if you do not want"
                " include_glob settings to be evaluated"
                " after the pre script is run. The"
                " default is 1."
            ),
            "enabled": __(
                "Set this to 0 if you want to disable all clients. The"
                " default is 1. This option can be overridden"
                " per-client in the client configuration files in"
                " clientconfdir on the server."
            ),
            "cname_fqdn": __(
                "Whether to keep fqdn cname (like"
                " 'testclient.example.com') when looking-up in"
                " clientconfdir. This also affects the fqdn lookup"
                " on the client (see client configuration options"
                " for details). The default is 1. When set to 0, the"
                " fqdn provided by the client while authenticating"
                " will be stripped ('testclient.example.com'"
                " becomes 'testclient')."
            ),
            "cname_lowercase": __(
                "Whether to force lowercase cname when"
                " looking-up in clientconfdir. This also"
                " affects the fqdn lookup on the client (see"
                " client configuration options for details)."
                " The default is 0. When set to 1 the name"
                " provided by the client while authenticating"
                " will be lowercased."
            ),
            "port": __(
                "Defines the main TCP port that the server listens on. "
                "Specify multiple 'port' entries on separate lines in "
                "order to listen on multiple ports. Each port can be "
                "configured with its own 'max_children' value."
            ),
            "max_children": __(
                "Defines the number of child processes to fork "
                "(the number of clients that can simultaneously "
                "connect. The default is 5. Specify multiple "
                "'max_children' entries on separate lines if you "
                "have configured multiple port entries."
            ),
            "status_port": __(
                "Defines the TCP port that the server listens on "
                "for status requests. Comment this out to have no "
                "status server. Specify multiple 'status_port' "
                "entries on separate lines in order to listen on "
                "multiple ports. Each port can be configured with "
                "its own 'max_status_children' value."
            ),
            "max_status_children": __(
                "Defines the number of status child "
                "processes to fork (the number of status "
                "clients that can simultaneously connect. "
                "The default is 5. Specify multiple "
                "'max_status_children' entries on separate "
                "lines if you have configured multiple "
                "status_port entries."
            ),
            "rblk_memory_max": __(
                "The maximum amount of data from the disk "
                "cached in server memory during a protocol2 "
                "restore/verify. The default is 256Mb. This "
                "option can be overriden per-client in the "
                "client configuration files in clientconfdir "
                "on the server."
            ),
        }
    )



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.custom.resource
    :platform: Unix
    :synopsis: Burp-UI api custom resource module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import inspect
import json

from flask_restx import Resource as ResourcePlus
from flask_restx.errors import abort

from ...tools.logging import logger


class Resource(ResourcePlus):
    """Subclass default Resource to manage ACL"""

    # by default, we require valid login for all routes. This can be disabled
    # per resource
    login_required = True

    logger = logger

    def abort(self, code=500, message=None, **kwargs):
        """
        Properly abort the current request

        See: :func:`~flask_restx.errors.abort`
        """
        if message and not isinstance(message, str):
            try:
                message = json.dumps(message)  # pragma: no cover
            except:
                message = None
        # Add extra logs when raising abort exception
        (frm, filename, line_no, func, source_code, source_index) = inspect.stack()[1]
        mod = inspect.getmodule(frm)
        self.logger.debug("Abort in {}:{}".format(filename, line_no))
        self.logger.warning(
            "[{}] {}: {}{}".format(
                mod.__name__, code, message, " - {}".format(kwargs) if kwargs else ""
            )
        )
        # This raises a Flask Exception
        abort(code, message, **kwargs)
